import numpy as np

def check_pidx_duplicates(pidx):
    """
    Given a ring match generted set of indices, apply outlier detection to
    ensure no duplicates exist in either the reference column or the source
    column. If duplicates do exist, remove all rows because the solution is ambiguous.
    """
    # Check for duplicates
    l = pidx[:,1].tolist()
    clean = [i for i, x in enumerate(l) if l.count(x) == 1]
    pidx = pidx[clean, :]

    l = pidx[:,0].tolist()
    clean = [i for i, x in enumerate(l) if l.count(x) == 1]
    pidx = pidx[clean, :]
    return pidx

def ransac_permute(ref_points, tar_points, tolerance_val, target_points):
    """
    Given a set of reference points and target points, compute the
    geometric distances between pairs of points in the reference set
    and pairs of points in the target set.  Points for which the ratio of
    the distances is within plus or minus 1 - tolerance_value are considered
    to be good matches.

    If a valid solution is not found, this func returns three empty lists.

    Parameters
    ----------
    ref_points : ndarray
                 A (n, 2) array of points where the first column is the x
                 pixel location and the second column is the y pixel location.
                 Additional columns are ignored. These points are from one
                 image.

    tar_points : ndarray
                 A (n,2) array as above, for a second image.

    tolerance_value : float
                      On the range [-1, 1], the computed ratio must be
                      within 1 +- tolerance to be considered a match

    target_points : int
                    The minimum number of points required to return
                    a valid answer

    Returns
    -------
    ref_points : ndarray
                 (n,2) subset of the input ref_points

    tar_points : ndarray
                 (n,2) subset of the input tar points

    f2 : ndarray
         of indices for valid points

    References
    ----------
    P. Sidiropoulos and J.-P. Muller, A systematic solution to multi-instrument co-registration of high-resolution planetary images to an orthorectified baseline, IEEE Transactions on Geoscience and Remote Sensing, 2017
    """
    n = len(ref_points)
    # Build an n,n distance matrix to store pairwise point distances
    dist = np.zeros((n,n))
    for i in range(n):
        vr1 = ref_points[i]
        vt1 = tar_points[i]
        for j in range(i, n):
            # These are diagonals so always zero
            if i == j:
                dist[i,j] = 0
                dist[j,i] = 0
                continue
            # Compute the distance between ref_b - ref_a and tar_b - tar_a. The
            # absolute value should be small as these points should be sp
            vr2 = ref_points[j]
            vt2 = tar_points[j]

            dr = vr2 - vr1
            dt = vt2 - vt1

            dist[i,j] = (dr[0]**2 + dr[1]**2)**0.5 / (dt[0]**2+dt[1]**2)**0.5
            dist[j,i] = dist[i,j]
    # Compute min/max bounds for the tolerance
    minlim = 1 - tolerance_val
    maxlim = 1 + tolerance_val

    # Determine which points are within the tolerance
    q1 = dist >= minlim
    q2 = dist <= maxlim
    q = (q1*q2).astype(np.int)
    # How many points are within the tolerance?
    s = np.sum(q, axis=1)
    # If the number of points within the tolerance are greater than the number of desired points
    if np.max(s) >= target_points:
        m = np.eye(n).dot(target_points + 1)
        for i in range(n):
            for j in range(i):
                m[i,j] = q[i].dot(q[j])
                m[j,i] = m[i,j]
        qm = m >= target_points
        sqm = np.sum(qm, axis=-1)
        f = np.argmax(sqm)
        f2 = np.nonzero(qm[f])
        return ref_points[f2], tar_points[f2], f2
    else:
        return [], [], []

def sift_match(a, b, thresh=1.5):
    """
    vl_ubcmatch from the vlfeat toolbox for MatLab.  This is
    Lowe's prescribed implementation for disambiguating descriptors.

    Parameters
    ----------
    a : np.ndarray
        (m,) a singular descriptors where the m-dimension are the
        descriptor lengths.  For SIFT m=128. This is reshaped from
        a vector to an array.
    b : np.ndarray
        (n,m) where the n-dimension are the individual features and
        the m-dimension are the elements of the descriptor.

    thresh : float
             The threshold for disambiguating correspondences. From Lowe.
             If best * thresh < second_best, a match has been found.

    Returns
    -------
    best : int
           Index for the best match

    References
    ----------
    P. Sidiropoulos and J.-P. Muller, A systematic solution to multi-instrument co-registration of high-resolution planetary images to an orthorectified baseline, IEEE Transactions on Geoscience and Remote Sensing, 2017
    """
    a = a.reshape(1,-1)
    dists = np.sum((b-a)**2, axis=1)

    try:
        best = np.nanargmin(dists)
    except:
        return

    if len(dists[dists != dists[best]]) == 0:
        return  # Edge case where all descriptors are the same
    elif len(dists[dists == dists[best]]) > 1:
        return  # Edge case where the best is ambiguous
    sec_best = np.nanmin(dists[dists != dists[best]])

    if dists[best] * thresh < sec_best:
        return best
    return

def ring_match(ref_feats, tar_feats, ref_desc, tar_desc, ring_radius=4000,
               max_radius=40000, target_points=15, tolerance_val=0.02,
               iteration_break_point=200):
    """
    Apply the University College London ring matching technique that seeks to match
    target feats to a number of reference features.

    Parameters
    ----------
    ref_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the target features

    tar_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the reference features

    ref_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    tar_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    ring_radius : numeric
                  The width of a ring for matching. In the same units as the x,y
                  coordinates for the features, e.g. if the ref_feats and tar_feats
                  are provided in pixel space and meters, the ring_radius should
                  be expressed in meters

    max_radius : numeric
                 The maximum radius to be tested.  This is the maxixum distance
                 a given correspondence could be from the initial estimate.

    target_points : int
                    The number of points that constitute a valid match

    tolerance : float
                The tolerance for outlier detection in point ordering between estimated
                resolutions

    Returns
    -------
    xref : ndarray
           (n,4) array of the correspondences selected from the ref_feats input

    xtar : ndarray
           (n,4) array of the correspondences selected from the tar_feats input

    p_idx : ndarray
            (n,2) array of the indices in the ref_feats and tar_feats input arrays

    References
    ----------
    P. Sidiropoulos and J.-P. Muller, A systematic solution to multi-instrument co-registration of high-resolution planetary images to an orthorectified baseline, IEEE Transactions on Geoscience and Remote Sensing, 2017
    """

    # Reference and target features
    ref_xy = ref_feats[:,:2]
    ref_xmym = ref_feats[:,2:4]
    tar_xy = tar_feats[:,:2]
    tar_xmym = tar_feats[:,2:4]
    
    # Boolean mask for those reference points that have already been matched
    ref_mask = np.ones(len(ref_xy), dtype=bool)
 
    # Counters
    numr = len(ref_feats)
    rad_num = int(max_radius / ring_radius) # Number of radial rings
    points_num = np.zeros(rad_num, dtype=np.int)  # Number of points per ring vector
    metr = 1
    
    # Initial array for holding candidate points
    p = np.zeros((target_points, 4 * rad_num))
    p_idx = np.zeros((target_points, 2 * rad_num), dtype=np.int)

    while ref_mask.any():
        # Grab a random reference point
        r = np.random.choice(np.arange(numr)[ref_mask])
        current_ref_desc = ref_desc[r]
        current_ref_xy = ref_xy[r]
        current_ref_xmym = ref_xmym[r]
        # Compute the euclidean distance between the reference point and all targets
        d = np.linalg.norm(current_ref_xmym - tar_xmym, axis=1)
        # For each point, independently match to a point in a given ring
        for i in range(rad_num):
            # The number of points that are within a given ring
            z = points_in_ring(d, i*ring_radius, (i+1) * ring_radius)
            # If we have enough points, run the sift matcher and select the best point, updating p
            if np.sum(z) >= target_points:
                # All candidate points that are in the ring
                current_tar_descs = tar_desc[z]  # This slicing uses ~25% of processing timr
                current_tar_xys = tar_xy[z]
                z_idx = np.where(z == True)[0]
                # Sift Match
                match = sift_match(current_ref_desc, current_tar_descs, thresh=1.5)  # The remaining 75% of processing time.
                if match is not None:
                    if points_num[i] == p.shape[0]:
                        p = dynamically_grow_array(p, target_points)
                        p_idx = dynamically_grow_array(p_idx, target_points, dtype=np.int)
                    p[points_num[i], 4*i:4*i+4] = [current_ref_xy[0], current_ref_xy[1], current_tar_xys[match][0], current_tar_xys[match][1]]
                    # Set the id of the point
                    p_idx[points_num[i], 2*i:2*i+2] = [r, z_idx[match]]
                    points_num[i] += 1
        
        if (metr % iteration_break_point == 0) or (np.sum(ref_mask) == 0):
            max_cons = 3
            # Find all candidate rings
            candidate_rings = points_num >= target_points
            for j in range(rad_num):
                if candidate_rings[j]:
                    # For each ring that is a candidate, select all of the reference and targets from the p matrix
                    # the first part of the slice grabs all candidate points and the second half in the first two args
                    # selects the ref and target (respectively).
                    npoints_in_ring = points_num[j]
                    ref_points = p[:npoints_in_ring, 4*j:4*j+2] # Slice out the reference coords
                    tar_points = p[:npoints_in_ring, 4*j+2:4*j+4]  # Slice out the target coords
                    indices = p_idx[:npoints_in_ring, 2*j:2*j+2] # slice out the indices

                    xref, xtar, idx = ransac_permute(ref_points, tar_points, tolerance_val, target_points)
                    # This selects the best of the rings
                    max_cons = max(max_cons, len(xref))
                    if len(xref) >= target_points:
                        # Solution found
                        # Instead of returning the ref and tar points, return the ids
                        ring = (j*ring_radius, j*ring_radius+ring_radius)
                        return xref, xtar, indices[idx], ring
        # Mask the reference point and iterate to the next randomly selected point
        ref_mask[r] = False
        metr += 1
    return None, None, None, 'Exhausted'

def points_in_ring(distance_vector, inner_radius, outer_radius):
    """
    Returns the indices of all points within a given distance from
    a distance vector. The vector is assumed to be 1d.

    Parameters
    ----------
    distance_vector : ndarray
                      (n,1) array of distances between a point and a set of points
    inner_radius : float
                   The lower threshold for the ring
    outer_radius : float
                   The outer threshold for the ring

    Returns
    -------
     : ndarray
       (n,1) array of booleans where all correspondences inside of the array are True
    """
    return (distance_vector >= inner_radius) * (distance_vector <= outer_radius)

def dynamically_grow_array(array, m, dtype=None):
    """
    Given an array, dynamically grow the array vertically with an m,n array of zeros

    Parameters
    ----------
    array : ndarray
            The array to be grown
    m : int
        The number of new rows to add

    dtype : obj
            A numpy data type that is used for the new entries. A 
            dynamically grown array will upcast to the most complex
            data type.

    Returns
    -------
    array : ndarray
            The expanded array
    """
    array_append = np.zeros((m, array.shape[1]), dtype=dtype)
    array = np.vstack((array, array_append))
    return array


def directed_ring_match(ref_feats, tar_feats, ref_desc, tar_desc, ring_min, ring_max, target_points=15, tolerance_value=0.02):
    """
    Given an input set of reference features and target features, attempt to find
    correspondences within a given ring, where the ring is defined by min and max
    radii.  This is a directed version of the ring_match function, in that this
    function assumes that the correspondence is within the defined ring.

    This implementation is inspired by and uses the ring_match implementation above, developed
    by Sidiropoulos and Muller.

    Parameters
    ----------
    ref_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the target features

    tar_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the reference features

    ref_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    tar_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    ring_min : numeric
               The inner distance of the ring

    ring_max : numeric
               The outer distance of the ring


    target_points : int
                    The number of points that constitute a valid match

    tolerance : float
                The tolerance for outlier detection in point ordering between estimated
                resolutions

    Returns
    -------
    xref : ndarray
           (n,4) array of the correspondences selected from the ref_feats input

    xtar : ndarray
           (n,4) array of the correspondences selected from the tar_feats input

    p_idx : ndarray
            (n,2) array of the indices in the ref_feats and tar_feats input arrays
    """

    # Reference and target features
    ref_xy = ref_feats[:,:2]
    ref_xmym = ref_feats[:,3:]
    tar_xy = tar_feats[:,:2]
    tar_xmym = tar_feats[:,3:]

    numr = len(ref_feats)

    p = np.zeros((numr, 4))
    p_idx = np.zeros((numr, 2), dtype=np.int)
    points_num = 0
    # Iterate over all of the reference points seeking a match
    for r in range(numr):
        current_ref_desc = ref_desc[r]
        current_ref_xy = ref_xy[r]
        current_ref_xmym = ref_xmym[r]

        # Compute the euclidean distance between the reference point and all targets
        d = np.linalg.norm(current_ref_xmym - tar_xmym, axis=1)
        # Find all of the candidates, if none, skip
        z = (d > ring_min) * (d < ring_max)

        if np.sum(z) == 0:
            continue
        current_tar_descs = tar_desc[z]
        current_tar_xys = tar_xy[z]
        z_idx = np.where(z == True)[0]

        match = sift_match(current_ref_desc, current_tar_descs, thresh=1.5)
        if match is not None:
            p[points_num] = [current_ref_xy[0], current_ref_xy[1], current_tar_xys[match][0], current_tar_xys[match][1]]
            # Set the id of the point
            p_idx[points_num] = [r, z_idx[match]]
            points_num += 1
    if points_num == 0:
        # No candidate matches found in this set.
        return [], [], []
    # Now that the candidates have all been located, check their geometric relationships to find the good matches
    ref_points = p[:points_num, :2]
    tar_points = p[:points_num, 2:]
    xref, xtar, idx = ransac_permute(ref_points, tar_points, tolerance_value, target_points)
    return xref, xtar, p_idx[idx]

def ring_match_one(x, y, ref_feats, tar_feats, ref_desc, tar_desc, ring, search_radius=600, max_search_radius=600, target_points=5):
    """
    Given an x,y coordinate where a match is desired, find all candidates within
    some search radius and attempt to generate a match.  If no matches are identified
    expand the search radius by search_radius and search again. Once a match
    is identified (by calling the directed matcher and finding a geometric consensus
    with at least target_points), the match closest to the given x,y is returned.

    Parameters
    ----------
    x : int
        x coordinate in image space

    y : int
        y coordinate in image space

    ref_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the target features

    tar_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the reference features

    ref_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    tar_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    ring : tuple
           in the form (min_ring, max_ring)

    search_radius : numeric
                    The radius of the search extent in image pixels

    target_points : int
                    The desired number of points to identify a correspondence
    """
    search = True

    # Start search within search_radius of the center of the cell
    while search_radius <= max_search_radius:
        # Find all correspondences within 100m of the center of the cell to fill
        candidates = np.where((np.abs(ref_feats[:,0]-y) < search_radius) &\
                              (np.abs(ref_feats[:,1]-x) < search_radius))[0]

        sref_feats = ref_feats[candidates]
        sref_desc = ref_desc[candidates]

        min_ring = ring[0]
        max_ring = ring[1]

        sx_ref, sx_tar, sindices, = directed_ring_match(sref_feats, tar_feats, sref_desc, tar_desc, min_ring, max_ring, target_points=target_points)
        if len(sindices) >= target_points:

            #sindices[:,0] = candidates[sindices[:,0]]

            # Find the match that is closest to the center of the cell.
            center = np.array([x, y])
            ref_coords = sref_feats[sindices[:,0], :2]
            dist = np.linalg.norm(center - ref_coords, axis=1)
            closest_idx = np.argmin(dist)

            # Remap the reference from the subset to the full set
            sindices[:,0] = candidates[sindices[:,0]]

            return sindices[closest_idx]

        # expand the search radius by 100m
        search_radius += search_radius
        if search_radius >= max_search_radius:
            return []

def add_correspondences(in_feats, ref_feats, tar_feats, ref_desc, tar_desc,  xextent, yextent, ring, n_x_cells=5, n_y_cells=5, target_points=5, **kwargs):
    """
    Given a set of input features and x/y extents lay a regular grid over the area
    defined by the extents and find correspondences within each grid cell that
    does not have any existing correspondences.

    Then number of cells are defined by the n_x/y_cells parameters.

    Parameters
    ----------
    in_feats : ndarray
               (n,m) input features with the first column being x-coordinate
               in image space and the second column being y-coordinate.

    ref_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the target features

    tar_feats : np.ndarray
                where the first 2 columns are the x,y coordinates in pixel space,
                columns 3 & 4 are the x,y coordinates in m or km in the same
                reference as the reference features

    ref_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    tar_desc : np.ndarray
               (m, n) where m are the individual SIFT features and
               n are the descriptor length (usually 128)

    xextent : tuple
              in the form (minx, maxx)

    yextent : tuple
              in the form (miny, maxy)

    ring : tuple
           in the form (min_ring, max_ring)

    n_x_cells : int
                the number of cells to generate in the x direction

    n_y_cells : int
                the number of cells to generate in the y direction

    target_points : int
                    The desired number of points to identify a correspondence

    See Also
    --------
    autocnet.matcher.cpu_ring_matcher: for options associated with kwargs parameter
    """
    x_edges = np.linspace(xextent[0], xextent[1], n_x_cells)
    y_edges = np.linspace(yextent[0], yextent[1], n_y_cells)

    # Find the cells that are populated and assign as covered
    xbins = np.digitize(in_feats[:,1], bins=x_edges)
    ybins = np.digitize(in_feats[:,0], bins=y_edges)
    covered = list(zip(xbins, ybins))

    refs_to_add = []
    # Loop over all cells
    for i in range(1, n_x_cells):
        for j in range(1, n_y_cells):
            # and process only the uncovered cells
            if (i,j) in covered:
                continue
            x_cell_center = x_edges[i-1] + (x_edges[i] - x_edges[i-1])/2
            y_cell_center = y_edges[j-1] + (y_edges[j] - y_edges[j-1])/2

            ref = ring_match_one(x_cell_center, y_cell_center, ref_feats, tar_feats, ref_desc, tar_desc, ring, **kwargs)
            refs_to_add.append(ref)
    return refs_to_add



import unittest
from unittest.mock import Mock, MagicMock
import pytest

import ogr
import numpy as np
import pandas as pd
from plio.io import io_gdal
from shapely.geometry import Polygon as Poly

from autocnet.matcher import cpu_outlier_detector as od
from autocnet.examples import get_path
from autocnet.graph.network import CandidateGraph
from autocnet.utils.utils import array_to_poly

from .. import edge
from .. import node

import logging
log = logging.getLogger(__name__)

class TestEdge(unittest.TestCase):

    def setUp(self):
        source = Mock(node.Node)
        destination = Mock(node.Node)
        self.edge = edge.Edge(source=source, destination=destination)

    def test_masks(self):
        self.assertIsInstance(self.edge.masks, pd.DataFrame)

    def test_edge_overlap(self):
        e = edge.Edge()
        source = Mock(spec = node.Node)
        destination = Mock(spec = node.Node)
        e.destination = destination
        e.source = source
        geodata_s = Mock(spec = io_gdal.GeoDataset)
        geodata_d = Mock(spec = io_gdal.GeoDataset)
        source.geodata = geodata_s
        destination.geodata = geodata_d

        wkt1 = "POLYGON ((0 40, 40 40, 40 0, 0 0, 0 40))"
        wkt2 = "POLYGON ((20 60, 60 60, 60 20, 20 20, 20 60))"

        poly1 = ogr.CreateGeometryFromWkt(wkt1)
        poly2 = ogr.CreateGeometryFromWkt(wkt2)

        source.geodata.footprint = poly1
        destination.geodata.footprint = poly2

        e.overlap()
        self.assertEqual(e.weights['overlap_area'], 400)
        self.assertAlmostEqual(e.weights['overlap_percn'], 14.285714285)

    def test_weight(self):
        e = edge.Edge()
        assert e.weights == {}
        e.weights = ('foo', 1)
        assert e.weights['foo']  == 1
        e.weights = ('foo', 2)
        assert e.weights['foo']  == 2

    def test_coverage(self):
        adjacency = get_path('two_image_adjacency.json')
        basepath = get_path('Apollo15')
        cg = CandidateGraph.from_adjacency(adjacency, basepath=basepath)
        keypoint_df = pd.DataFrame({'x': (15, 18, 18, 12, 12), 'y': (5, 10, 15, 15, 10)})
        keypoint_matches = [[0, 0, 1, 0],
                            [0, 1, 1, 1],
                            [0, 2, 1, 2],
                            [0, 3, 1, 3],
                            [0, 4, 1, 4]]

        matches_df = pd.DataFrame(data=keypoint_matches, columns=['source_image', 'source_idx', 'destination_image', 'destination_idx'])

        e = edge.Edge()
        source_node = MagicMock(spec=node.Node())
        destination_node = MagicMock(spec=node.Node())

        source_node.get_keypoint_coordinates = MagicMock(return_value=keypoint_df)
        destination_node.get_keypoint_coordinates = MagicMock(return_value=keypoint_df)

        e.source = source_node
        e.destination = destination_node

        source_geodata = Mock(spec=io_gdal.GeoDataset)
        destination_geodata = Mock(spec=io_gdal.GeoDataset)

        e.source.geodata = source_geodata
        e.destination.geodata = destination_geodata

        source_corners = [(0, 0),
                          (20, 0),
                          (20, 20),
                          (0, 20)]

        destination_corners = [(10, 5),
                               (30, 5),
                               (30, 25),
                               (10, 25)]

        e.source.geodata.latlon_corners = source_corners
        e.destination.geodata.latlon_corners = destination_corners

        vals = {(15, 5): (15, 5), (18, 10): (18, 10), (18, 15): (18, 15), (12, 15): (12, 15), (12, 10): (12, 10)}

        def pixel_to_latlon(i, j):
            return vals[(i, j)]

        e.source.geodata.pixel_to_latlon = MagicMock(side_effect=pixel_to_latlon)
        e.destination.geodata.pixel_to_latlon = MagicMock(side_effect=pixel_to_latlon)

        e.matches = matches_df

        #self.assertRaises(AttributeError, cg.edge[0][1].coverage)
        self.assertEqual(e.coverage(), 0.3)

    def test_get_keypoints(self):
        src_keypoint_df = pd.DataFrame({'x': (0, 1, 2, 3, 4), 'y': (5, 6, 7, 8, 9),
                                        'response': (10, 11, 12, 13, 14), 'size': (15, 16, 17, 18, 19),
                                        'angle': (20, 21, 22, 23, 24), 'octave': (25, 26, 27, 28, 29),
                                        'layer': (30, 31, 32, 33, 34)})

        dst_keypoint_df = pd.DataFrame({'x': (34, 33, 32, 31, 30), 'y': (29, 28, 27, 26, 25),
                                        'response': (24, 23, 22, 21, 20), 'size': (19, 18, 17, 16, 15),
                                        'angle': (14, 13, 12, 11, 10), 'octave': (9, 8, 7, 6, 5),
                                        'layer': (4, 3, 2, 1, 0)})

        keypoint_matches = [[0, 0, 1, 4],
                            [0, 1, 1, 3],
                            [0, 2, 1, 2],
                            [0, 3, 1, 1],
                            [0, 4, 1, 0]]

        matches_df = pd.DataFrame(data=keypoint_matches, columns=['source_image', 'source_idx',
                                                                  'destination_image', 'destination_idx'])
        source_node = node.Node(node_id=0)
        destination_node = node.Node(node_id=1)
        e = edge.Edge(source_node, destination_node)

        source_node.get_keypoint_coordinates = MagicMock(return_value=src_keypoint_df)
        destination_node.get_keypoint_coordinates = MagicMock(return_value=dst_keypoint_df)

        e.clean = MagicMock(return_value=(matches_df, None))
        e.matches = matches_df

        # Test all uses for edge.get_keypoints()
        src_matched_keypts = e.get_keypoints("source")
        src_matched_keypts2 = e.get_keypoints(e.source)
        dst_matched_keypts = e.get_keypoints("destination")
        dst_matched_keypts2 = e.get_keypoints(e.destination)

        # [output df to test] [name of node] [df to test against]
        to_test = [[src_matched_keypts, "source", src_keypoint_df],
                   [src_matched_keypts2, "source", src_keypoint_df],
                   [dst_matched_keypts, "destination", dst_keypoint_df],
                   [dst_matched_keypts2, "destination", dst_keypoint_df]]

        for out_df in to_test:
            # For each row index in the appropriate column of the matches_df,
            # assert that row index exists in the function's returned df
            [self.assertIn(row_idx, out_df[0].index.values)
             for row_idx in matches_df[out_df[1] + '_idx']]
            # For each row index in the returned df
            for row_idx in out_df[0].index.values:
                # Assert that row_idx exists in the matches_df's appropriate
                # column
                self.assertIn(row_idx, matches_df[out_df[1] + '_idx'])
                # Assert that all row_idx[column] vals returned by function
                # match their counterpart in orig df
                for column in out_df[0].columns:
                    self.assertTrue(out_df[0].iloc[row_idx][column] ==
                                            out_df[2].iloc[row_idx][column])

        # Test when overlap=True
        # edge["source_mbr"] and edge["destin_mbr"] haven't been calculated
        # yet, so return val should be unmasked df of node's keypts
        s_no_overlap = e.get_keypoints(e.source, overlap=True)
        d_no_overlap = e.get_keypoints(e.destination, overlap=True)
        self.assertTrue(s_no_overlap.equals(src_keypoint_df))
        self.assertTrue(d_no_overlap.equals(dst_keypoint_df))

        # Define the MBRs
        e.overlap_latlon_coords = 0, 0

        source_node.reproject_geom = MagicMock(return_value=Poly([(1, 6), (1, 8), (3, 6), (3, 8)]))
        source_node.keypoints = src_keypoint_df

        destination_node.reproject_geom = MagicMock(return_value=Poly([(31, 26), (31, 28), (33, 26), (33, 28)]))
        destination_node.keypoints = dst_keypoint_df

        # Only keep keypt vals w/I these bounding rects
        e["source_mbr"] = (0, 2, 5, 8)
        e["destin_mbr"] = (31, 33, 26, 28)

        # Grab the keypoints on our MBR overlaps
        s_overlap = e.get_keypoints(e.source, overlap=True)
        d_overlap = e.get_keypoints(e.destination, overlap=True)

        # Assert masked src keypts coords are equal
        s_expected = pd.DataFrame({'x': (0, 1, 2), 'y': (5, 6, 7)})
        self.assertTrue(np.array_equal(s_expected['x'], s_overlap['x'].values))
        self.assertTrue(np.array_equal(s_expected['y'], s_overlap['y'].values))

        # Assert masked dst keypt coords are equal
        d_expected = pd.DataFrame({'x': (33, 32, 31), 'y': (28, 27, 26)})
        self.assertTrue(np.array_equal(d_expected['x'], d_overlap['x'].values))
        self.assertTrue(np.array_equal(d_expected['y'], d_overlap['y'].values))

        # Assert type-checking in method throws proper errors
        with self.assertRaises(TypeError):
            e.get_keypoints("source", index = 456)
        with self.assertRaises(AttributeError):
            e.get_keypoints(1)
        # Check key error thrown when string arg != "source" or "destination"
        with self.assertRaises(AttributeError):
            e.get_keypoints("string")

    def test_eq(self):
        edge1 = edge.Edge()
        edge2 = edge.Edge()
        edge3 = edge.Edge()

        # Test edges w/ different keys are not equal, ones with same keys are
        edge1.__dict__["key"] = 1
        edge2.__dict__["key"] = 1
        edge3.__dict__["not_key"] = 1

        self.assertTrue(edge1 == edge2)
        self.assertFalse(edge1 == edge3)

        # Test edges with same keys, but diff df values
        edge1.__dict__["key"] = pd.DataFrame({'x': (0, 1, 2, 3, 4)})
        edge2.__dict__["key"] = pd.DataFrame({'x': (0, 1, 2, 3, 4)})
        edge3.__dict__["key"] = pd.DataFrame({'x': (0, 1, 2, 3, 5)})

        self.assertTrue(edge1 == edge2)
        self.assertFalse(edge1 == edge3)

        # Test edges with same keys, but diff np array vals
        # edge.__eq__ calls ndarray.all(), which checks that
        # all values in an array eval to true
        edge1.__dict__["key"] = np.array([True, True, True], dtype=bool)
        edge2.__dict__["key"] = np.array([True, True, True], dtype=bool)
        edge3.__dict__["key"] = np.array([True, True, False], dtype=bool)

        self.assertTrue(edge1 == edge2)
        self.assertFalse(edge1 == edge3)

    def test_repr(self):
        src = node.Node()
        dst = node.Node()
        masks = pd.DataFrame()

        e = edge.Edge()
        e.source = src
        e.destination = dst

        expected = """
        Source Image Index: {}
        Destination Image Index: {}
        Available Masks: {}
        """.format(src, dst, masks)

        self.assertEqual(expected, e.__repr__())

    @pytest.mark.filterwarnings('ignore::UserWarning')
    def test_ratio_check(self):
        """
        A pretty basic test that simply tests pass through from the edge to the
        proper distance func.
        """
        # Matches is init to None
        e = edge.Edge()

        # If there are matches...
        keypoint_matches = [[0, 0, 1, 4, 5],
                            [0, 1, 1, 3, 5],
                            [0, 2, 1, 2, 5],
                            [0, 3, 1, 1, 5],
                            [0, 4, 1, 0, 5]]

        matches_df = pd.DataFrame(data=keypoint_matches, columns=['source_image', 'source_idx',
                                                                  'destination_image', 'destination_idx', 'distance'])
        e.matches = matches_df
        expected = od.distance_ratio(None, matches_df)
        e.ratio_check()
        assert expected.equals(e.masks["ratio"])

    def test_overlap_check(self):
        s = node.Node(node_id=0)
        d = node.Node(node_id=1)

        e = edge.Edge(s, d)

        src_keypoint_df = pd.DataFrame({'x': (0, 1, 2, 3, 4), 'y': (5, 6, 7, 8, 9)})
        dst_keypoint_df = pd.DataFrame({'x': (34, 33, 32, 31, 30), 'y': (29, 28, 27, 26, 25)})

        # Create keypt matches
        keypoint_matches = [[0, 0, 1, 4, 5],
                            [0, 1, 1, 3, 5],
                            [0, 2, 1, 2, 5],
                            [0, 3, 1, 1, 5],
                            [0, 4, 1, 0, 5]]

        matches_df = pd.DataFrame(data=keypoint_matches,
                                  columns=['source_image', 'source_idx',
                                           'destination_image', 'destination_idx', 'distance'])
        s.keypoints = src_keypoint_df
        d.keypoints = dst_keypoint_df
        e.matches = matches_df

        s_overlap_keypts = pd.DataFrame({'x': (0, 1), 'y': (5, 6)})
        d_overlap_keypts = pd.DataFrame({'x': (31, 30), 'y': (26, 25)}, index=[3, 4])
        expected_mask = pd.Series(data=[True, True, False, False, False])

        # Mockup of the Edge.get_keypoints() method when overlap=True
        def mock_get_keypts(node, overlap=False):
            if node == s and overlap:
                return s_overlap_keypts
            elif node == d and overlap:
                return d_overlap_keypts
            else:
                return None

        e.get_keypoints = MagicMock(side_effect=mock_get_keypts)

        # Should fail if no src & dst mbrs on edge; Warns user & mask isn't
        # populated
        with self.assertLogs() as captured:
            e.overlap_check()
        self.assertTrue("Overlap between" not in captured.records[0].getMessage()) 

        # Should work after MBRs are set
        e["source_mbr"] = (1, 1, 1, 1)
        e["destin_mbr"] = (1, 1, 1, 1)
        e.overlap_check()
        overlap_matches, overlap_mask = e.clean(clean_keys=['overlap'])
        self.assertTrue(expected_mask.equals(overlap_mask))
        self.assertTrue(overlap_matches.equals(e.matches[overlap_mask]))

    def test_bad_matches_type(self):
        with self.assertRaises(TypeError):
            s = node.Node(node_id=0)
            d = node.Node(node_id=1)

            e = edge.Edge(s, d)
            e.matches = ['a', 'b', 'c']

    def test_prep_subpixel_1(self):
        nmatches = 10
        nstrengths = 1
        e = edge.Edge()
        arrs = e._prep_subpixel(nmatches, nstrengths=nstrengths)
        assert len(arrs) == 5
        assert arrs[2].shape == (nmatches, nstrengths)
        assert arrs[0][0] == 0

    def test_prep_subpixel_2(self):
        nmatches = 10
        nstrengths = 2
        e = edge.Edge()
        arrs = e._prep_subpixel(nmatches, nstrengths=nstrengths)
        assert len(arrs) == 5
        assert arrs[2].shape == (nmatches, nstrengths)
        assert arrs[0][0] == 0



import warnings

from cv2 import ORB_create, FastFeatureDetector_create
import numpy as np
import pandas as pd
from autocnet.utils.utils import bytescale

try:
    import cyvlfeat as vl
    vlfeat = True
except Exception:  # pragma: no cover
    vlfeat = False
    pass

try:
    from cv2 import xfeatures2d
    SIFT = xfeatures2d.SIFT_create
    SURF = xfeatures2d.SURF_create
except Exception:  # pragma: no cover
    SIFT = None
    SURF = None
    pass


def extract_features(array, extractor_method='sift', extractor_parameters={}):
    """
    This method finds and extracts features from an image using the given dictionary of keyword arguments.
    The input image is represented as NumPy array and the output features are represented as keypoint IDs
    with corresponding descriptors.

    Parameters
    ----------
    array : ndarray
            a NumPy array that represents an image

    extractor_method : {'orb', 'sift', 'fast', 'surf', 'vl_sift'}
              The detector method to be used.  Note that vl_sift requires that
              vlfeat and cyvlfeat dependencies be installed.

    extractor_parameters : dict
                           A dictionary containing OpenCV SIFT parameters names and values.

    Returns
    -------
    keypoints : DataFrame
                data frame of coordinates ('x', 'y', 'size', 'angle', and other available information)

    descriptors : ndarray
                  Of descriptors
    """
    detectors = {'fast': FastFeatureDetector_create,
                 'sift': SIFT,
                 'surf': SURF,
                 'orb' : ORB_create}

    if extractor_method == 'vlfeat' and vlfeat != True:
        raise ImportError('VLFeat is not available.  Please install vlfeat or use a different extractor.')

    if  extractor_method == 'vlfeat':
        keypoint_objs, descriptors  = vl.sift.sift(array,
                                                   compute_descriptor=True,
                                                   float_descriptors=True,
                                                   **extractor_parameters)
        # Swap columns for value style access, vl_feat returns y, x
        keypoint_objs[:, 0], keypoint_objs[:, 1] = keypoint_objs[:, 1], keypoint_objs[:, 0].copy()
        keypoints = pd.DataFrame(keypoint_objs, columns=['x', 'y', 'size', 'angle'])
    else:
        # OpenCV requires the input images to be 8-bit
        if not array.dtype == 'int8':
            array = bytescale(array)
        detector = detectors[extractor_method](**extractor_parameters)
        keypoint_objs, descriptors = detector.detectAndCompute(array, None)

        keypoints = np.empty((len(keypoint_objs), 7), dtype=np.float32)
        for i, kpt in enumerate(keypoint_objs):
            octave = kpt.octave & 8
            layer = (kpt.octave >> 8) & 255
            if octave < 128:
                octave = octave
            else:
                octave = (-128 | octave)
            keypoints[i] = kpt.pt[0], kpt.pt[1], kpt.response, kpt.size, kpt.angle, octave, layer  # y, x
        keypoints = pd.DataFrame(keypoints, columns=['x', 'y', 'response', 'size',
                                                     'angle', 'octave', 'layer'])

        if hasattr(descriptors, 'dtype') and descriptors.dtype != np.float32:
            descriptors = descriptors.astype(np.float32)

    return keypoints, descriptors

def extract_most_interesting(image, extractor_method='orb', extractor_parameters={'nfeatures':10}):
    """
    Given an image, extract the most interesting feature. Interesting is defined
    as the feature descriptor that has the maximum variance. By default, this func
    finds 10 features in the image and then selects the best.

    Parameters
    ----------
    image : ndarray
            of DN values

    extractor_method : str
                       Any valid, autocnet extractor. Default (orb)

    exctractor_parameters : dict
                            of extractor parameters passed through to the feature extractor

    Returns
    -------
     : pd.series
       The keypoints row with the higest variance. The row has 'x' and 'y' columns to
       get the location.
    """
    kps, desc = extract_features(image,
                                 extractor_method=extractor_method,
                                 extractor_parameters=extractor_parameters)

    if len(kps) == 0:
        return None
    # Naively assume that the maximum variance is the most unique feature
    vari = np.var(desc, axis=1)
    return kps.iloc[np.argmax(vari)]



"""This module has tests for the change_detection functions."""

# This is free and unencumbered software released into the public domain.
#
# The authors of autocnet do not claim copyright on the contents of this file.
# For more details about the LICENSE terms and the AUTHORS, you will
# find files of those names at the top level of this repository.
#
# SPDX-License-Identifier: CC0-1.0

import contextlib
import unittest
from pathlib import Path
from subprocess import CalledProcessError
from unittest.mock import patch

from autocnet.utils import hirise

class TestISIS(unittest.TestCase):

    def setUp(self) -> None:
        self.resourcedir = Path("test-resources")
        red50img = self.resourcedir / "PSP_010502_2090_RED5_0.img"
        red51img = self.resourcedir / "PSP_010502_2090_RED5_1.img"

        if not all((red50img.exists(), red51img.exists())):
            self.skipTest(
                f"One or more files is missing from the {self.resourcedir} "
                "directory. Tests on real files skipped.")

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            for g in ("*.cub", "*.log"):
                for p in self.resourcedir.glob(g):
                    p.unlink()
            Path("print.prt").unlink()

    def test_ingest_hirise(self):
        hirise.ingest_hirise(str(self.resourcedir))
        self.assertTrue((
            self.resourcedir / "PSP_010502_2090_RED5.stitched.norm.cub"
        ).exists())

        with patch(
            "kalasiris.hi2isis",
                side_effect=CalledProcessError(returncode=1, cmd="foo")
        ):
            hirise.ingest_hirise(str(self.resourcedir))

    def test_segment_hirise(self):
        hirise.ingest_hirise(str(self.resourcedir))
        hirise.segment_hirise(str(self.resourcedir))
        base = self.resourcedir / "PSP_010502_2090_RED5.stitched.norm.cub"
        for f in (
            base.with_suffix(".1_1325.cub"),
            base.with_suffix(".725_2349.cub"),
            base.with_suffix(".1749_3373.cub"),
            base.with_suffix(".2773_4000.cub"),
        ):
            self.assertTrue(f.exists())


import os
import sys
import unittest

import numpy as np
import pandas as pd

sys.path.append(os.path.abspath('..'))
from .. import suppression_funcs as sf


class Test_Suppression(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        r = np.random.RandomState(12345)
        cls.df = pd.DataFrame(r.uniform(0, 1, (10, 4)), columns=['response',
                                                                 'correlation',
                                                                 'distance',
                                                                 'error'])

    def test_response(self):
        col = self.df.apply(sf.response, axis=1, args=[None])
        self.assertEqual(col.all(), self.df['response'].all())

    def test_correlation(self):
        col = self.df.apply(sf.correlation, axis=1, args=[None])
        self.assertEqual(col.all(), self.df['correlation'].all())

    def test_distance(self):
        col = self.df.apply(sf.distance, axis=1, args=[None])
        self.assertEqual(col.all(), self.df['distance'].all())

    def test_error(self):
        col = self.df.apply(sf.error, axis=1, args=[None])
        self.assertEqual(col.all(), self.df['error'].all())




import unittest
import numpy as np
import pandas as pd

from osgeo import ogr
from .. import utils


class TestUtils(unittest.TestCase):

    def setUp(self):
        pass

    def test_cross_form(self):
        a = np.array([-1, 0, 1.25])
        np.testing.assert_array_almost_equal(utils.crossform(a), np.array([[0., -1.25,  0.],
                                                                           [1.25,  0.,  1.],
                                                                           [-0., -1.,  0.]]))

    def test_checkbandnumbers(self):
        self.assertTrue(utils.checkbandnumbers([1,2,3,4,5], (2,5,1)))
        self.assertFalse(utils.checkbandnumbers([1,2,4], (1,2,3)))
        self.assertTrue(utils.checkbandnumbers([1.0, 2.0, 3.0], [1.0]))
        self.assertFalse(utils.checkbandnumbers([-1.0, 2.0, 3.0], (1.0, 2.0, 3.0)))

    def test_getdeplaid(self):
        self.assertEqual(utils.checkdeplaid(95), 'night')
        self.assertEqual(utils.checkdeplaid(127.4), 'night')
        self.assertEqual(utils.checkdeplaid(180), 'night')
        self.assertEqual(utils.checkdeplaid(94.99), 'night')
        self.assertEqual(utils.checkdeplaid(90), 'night')
        self.assertEqual(utils.checkdeplaid(26.1), 'day')
        self.assertEqual(utils.checkdeplaid(84.99), 'day')
        self.assertEqual(utils.checkdeplaid(0), 'day')
        self.assertFalse(utils.checkdeplaid(-1.0))

    def test_checkmonotonic(self):
        self.assertTrue(utils.checkmonotonic(np.arange(10)))
        self.assertTrue(utils.checkmonotonic(range(10)))
        self.assertFalse(utils.checkmonotonic([1,2,4,3]))
        self.assertFalse(utils.checkmonotonic([-2.0, 0.0, -3.0]))

        self.assertEqual(utils.checkmonotonic(np.arange(10), piecewise=True),
                [True] * 10)
        self.assertEqual(utils.checkmonotonic(range(10), piecewise=True),
                [True] * 10)
        self.assertEqual(utils.checkmonotonic([1,2,4,3], piecewise=True),
                [True,True,True, False])
        self.assertEqual(utils.checkmonotonic([-2.0, 0.0, -3.0],piecewise=True),
                [True,True,False])

    def test_getnearest(self):
        iterable = range(10)
        idx, value = utils.getnearest(iterable, 3)
        self.assertEqual(idx, 3)

        idx, value = utils.getnearest(iterable, 8.32)
        self.assertEqual(idx, 8)

        idx, value = utils.getnearest(iterable, 8.5)
        self.assertEqual(idx, 8)

        idx, value = utils.getnearest(iterable, 8.51)
        self.assertEqual(idx, 9)

    def test_find_in_dict(self):
        d = {'a':1,
            'b':2,
            'c':{
                'd':3,
                'e':4,
                'f':{
                    'g':5,
                    'h':6
                    }
                }
            }

        self.assertEqual(utils.find_in_dict(d, 'a'), 1)
        self.assertEqual(utils.find_in_dict(d, 'f'), {'g':5,'h':6})
        self.assertEqual(utils.find_in_dict(d, 'e'), 4)

    def test_find_nested_in_dict(self):
        d = {'a':1,
            'b':2,
            'c':{
                'd':3,
                'e':4,
                'f':{
                    'g':5,
                    'h':6
                    }
                }
            }

        self.assertEqual(utils.find_nested_in_dict(d, 'a'), 1)
        self.assertEqual(utils.find_nested_in_dict(d, ['c', 'f', 'g']), 5)

    def test_make_homogeneous(self):
        pts = np.arange(50).reshape(25,2)
        pts = utils.make_homogeneous(pts)
        self.assertEqual(pts.shape, (25,3))
        np.testing.assert_array_equal(pts[:, -1], np.ones(25))

    def test_remove_field_name(self):
        starray = np.array([(1 ,2.,'String'), (2, 3.,"String2")],
              dtype=[('index', 'i4'),('bar', 'f4'), ('baz', 'S10')])
        truth = np.array([(2.,'String'), (3.,"String2")],
              dtype=[('bar', 'f4'), ('baz', 'S10')])
        cleaned_array = utils.remove_field_name(starray, 'index')
        np.testing.assert_array_equal(cleaned_array, truth)

    def test_slope(self):
        x1 = pd.DataFrame({'x': np.arange(1, 11),
                           'y': np.arange(1, 11)})
        x2 = pd.DataFrame({'x': np.arange(6, 16),
                           'y': np.arange(11, 21)})

        slope = utils.calculate_slope(x1, x2)
        self.assertEqual(slope[0], 2)

    def test_array_to_poly(self):
        array1 = np.array([[1, 2],
                           [3, 4],
                           [5, 6]])
        array2 = np.array([[1, 2, 3],
                           [4, 5, 6],
                           [7, 8, 9]])
        geom1 = utils.array_to_poly(array1)

        self.assertIsInstance(geom1, ogr.Geometry)
        self.assertRaises(ValueError, utils.array_to_poly, array2)

    def test_dispatch(self):
        class Patchwork(object):

            def __init__(self, **kwargs):
                for k, v in kwargs.items():
                    setattr(self, k, v)

            @utils.methodispatch
            def get(self, arg):
                return getattr(self, arg, None)

            @get.register(list)
            def _(self, arg):
                return [self.get(x) for x in arg]

        patchwork = Patchwork(a=1, b=2, c=3)
        self.assertEqual(patchwork.get(['a', 'b']), [1, 2])
        self.assertEqual(patchwork.get('c'), 3)

    def test_decorate_class(self):
        class Test(object):
            def __init__(self):
                self.test = 'original'

            def get_test(self):
                return self.test

        def dec(func):
            return lambda x:'decorated'

        Dec_Test = utils.decorate_class(Test, dec)

        undecorated = Test()
        decorated = Dec_Test()

        self.assertEqual(undecorated.get_test(), 'original')
        self.assertEqual(decorated.get_test(), 'decorated')

        with self.assertRaises(Exception):
            utils.decorate_class(Test, 'Totally not a callable')

    def test_generate_decorator(self):
        def func_to_wrap(x):
            return x+1

        def wrapper():
            # Should be able to access run-time namespace
            return ret + 1, test

        decorator = utils.create_decorator(wrapper, test=0)
        wrapped_func = decorator(func_to_wrap)

        self.assertTrue(wrapped_func(1),2)



#!/usr/bin/env python

import argparse
import copy
import os
import json
import sys
import logging

from io import StringIO
from contextlib import redirect_stdout

from redis import StrictRedis

from autocnet.graph.network import NetworkCandidateGraph
from autocnet.graph.node import NetworkNode
from autocnet.graph.edge import NetworkEdge
from autocnet.io.db.model import Points, Measures, Overlay
from autocnet.utils.utils import import_func
from autocnet.utils.serializers import JsonEncoder, object_hook
from autocnet.io.db.model import JobsHistory

log = logging.getLogger(__name__)

def parse_args():  # pragma: no cover
    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--host', help='The host URL for the redis queue to to pull messages from.')
    parser.add_argument('-p', '--port', help='The port for used by redis.')
    parser.add_argument('-q', '--queue', default=False, action='store_true',
                        help='If passed, run in queue mode, where this job runs until either \
                              walltime is hit or the queue that is being processed is empty.')
    parser.add_argument('processing_queue', help='The name of the processing queue to draw messages from.')
    parser.add_argument('working_queue', help='The name of the queue to push messages to while they process.')

    return parser.parse_args()

def _instantiate_obj(msg, ncg):
    """
    Instantiate either a NetworkNode or a NetworkEdge that is the
    target of processing.

    """
    along = msg['along']
    id = msg['id']
    image_path = msg['image_path']
    if along == 'node':
        obj = NetworkNode(node_id=id, image_path=image_path)
    elif along == 'edge':
        obj = NetworkEdge()
        obj.source = NetworkNode(node_id=id[0], image_path=image_path[0])
        obj.destination = NetworkNode(node_id=id[1], image_path=image_path[1])
    obj.parent = ncg
    return obj

def _instantiate_row(msg, ncg):
    """
    Instantiate some db.io.model row object that is the target
    of processing.
    """
    # Get the dict mapping iterable keyword types to the objects
    objdict = ncg.apply_iterable_options
    obj = objdict[msg['along']]
    with ncg.session_scope() as session:
        res = session.query(obj).filter(getattr(obj, 'id')==msg['id']).one()
        session.expunge_all() # Disconnect the object from the session
    return res

def process(msg):
    """
    Given a message, instantiate the necessary processing objects and
    apply some generic function or method.

    Parameters
    ----------
    msg : dict
          The message that parametrizes the job.
    """
    ncg = NetworkCandidateGraph()
    ncg.config_from_dict(msg['config'])
    if msg['along'] in ['node', 'edge']:
        obj = _instantiate_obj(msg, ncg)
    elif msg['along'] in ['candidategroundpoints', 'points', 'measures', 'overlaps', 'images']:
        obj = _instantiate_row(msg, ncg)
    else:
        obj = msg['along']

    # Grab the function and apply. This assumes that the func is going to
    # have a True/False return value. Basically, all processing needs to
    # occur inside of the func, nothing belongs in here.
    #
    # All args/kwargs are passed through the RedisQueue, and then right on to the func.
    func = msg['func']
    if callable(func):  # The function is a de-serialzied function
        msg['args'] = (obj, *msg['args'])
        msg['kwargs']['ncg'] = ncg
    elif hasattr(obj, msg['func']):  # The function is a method on the object
        func = getattr(obj, msg['func'])
    else:  # The func is a function from a library to be imported
        func = import_func(msg['func'])
        # Get the object begin processed prepended into the args.
        msg['args'] = (obj, *msg['args'])
        # For now, pass all the potential config items through
        # most funcs will simply discard the unnecessary ones.
        msg['kwargs']['ncg'] = ncg
        msg['kwargs']['Session'] = ncg.Session

    # Now run the function.
    res = func(*msg['args'], **msg['kwargs'])

    # Update the message with the True/False
    msg['results'] = res
    # Update the message with the correct callback function

    return msg

def transfer_message_to_work_queue(queue, queue_from, queue_to):
    """
    Atomic pop/push from one redis list to another

    Parameters
    ----------
    queue : object
            PyRedis queue

    queue_from : str
                 The name of the queue to pop a message from

    queue_to : str
               The name of the queue to push a message to

    Returns
    -------
      : str
        The message from the queue
    """
    return queue.rpoplpush(queue_from, queue_to)

def finalize_message_from_work_queue(queue, queue_name, remove_key):
    """
    Remove a message from a queue

    Parameters
    ----------
    queue : object
            PyRedis queue

    queue_name : str
                 The name of the queue to remove an object from

    remove_key : obj
                 The message to remove from the list
    """
    # The operation completed. Remove this message from the working queue.
    queue.lrem(queue_name, 0, remove_key)

def manage_messages(args, queue):
    """
    This function manages pulling a message from a redis list, atomically pushing
    the message to another redis list, launching a generic processing job,
    and finalizing the message by removing it from the intermediary redis list.

    This function is an easily testable main for the cluster_submit CLI.

    Parameters
    ----------
    args : dict
           A dictionary with queue names that are parsed from the CLI

    queue : obj
            A py-Redis queue object

    """
    processing = True

    while processing:
        # Pop the message from the left queue and push to the right queue; atomic operation
        msg = transfer_message_to_work_queue(queue,
                                            args['processing_queue'],
                                            args['working_queue'])

        if msg is None:
            if args['queue'] == False:
                log.warning('Expected to process a cluster job, but the message queue is empty.')
                return
            elif args['queue'] == True:
                log.info(f'Completed processing from queue: {queue}.')
                return

        # The key to remove from the working queue is the message. Essentially, find this element
        # in the list where the element is the JSON representation of the message. Maybe swap to a hash?
        remove_key = msg

        #Convert the message from binary into a dict
        msgdict = json.loads(msg, object_hook=object_hook)

        # Apply the algorithm
        response = process(msgdict)
        log.info(response)

        #serializedDict = json.loads(msg)
        #results  = msgdict['results'] if msgdict['results'] else [{"status" : "success"}]
        #success = True if "success" in results[0]["status"].split(" ")[0].lower() else False

        #jh = JobsHistory(jobId=int(os.environ["SLURM_JOB_ID"]), functionName=msgdict["func"], args={"args" : serializedDict["args"], "kwargs": serializedDict["kwargs"]}, results=msgdict["results"], logs=out, success=success)

        #with response['kwargs']['Session']() as session:
            #session.add(jh)
            #session.commit()

        finalize_message_from_work_queue(queue, args['working_queue'], remove_key)

        # Process only a single job, else draw the next message off the queue if available.
        if args['queue'] == False:
            processing = False


def main():  # pragma: no cover
    args = vars(parse_args())
    # set up the logger
    logging.basicConfig(level=os.environ.get("AUTOCNET_LOGLEVEL", "INFO"))
    # Get the message
    queue = StrictRedis(host=args['host'], port=args['port'], db=0)
    manage_messages(args, queue)

if __name__ == '__main__':
    main()



import json
import time
import warnings

from sqlalchemy import insert
from sqlalchemy.sql.expression import bindparam

from autocnet.io.db.model import Points, Measures
from autocnet.utils.serializers import object_hook
from autocnet.transformation.spatial import reproject, og2oc

def watch_insert_queue(queue, queue_name, counter_name, engine, stop_event, sleep_time=5):
    """
    A worker process to be launched in a thread that will asynchronously insert or update 
    objects in the Session using dicts pulled from a redis queue. Using this queuing approach
    many cluster jobs are able to push to the redis queue rapidly and then a single writer
    process can push the data back to the database.
    
    This function requires that the function called by the asynchronous cluster job INCR
    (increment) the counter_name key in the redis cache. This counter is INCR (incremented) 
    by cluster jobs to track how many messages have been pushed to the queue (queue_name). 
    This func then reads that many messages and DECR (de-increments) the counter by that
    many messages. This way this function only reads when data is present and reads can occur 
    asynchronously. This works becase the cluster job pushes to the right side of the redis
    list and this function reads n-messages from the left side.
    
    This method uses the sqlalchemy core interface for performance reasons. Therefore, some
    mundging of column names is used to ensure that the model to be processed matches the
    database column names.
    
    Parameters
    ----------
    queue : obj
            A Redis or StrictRedis connection instance
    
    queue_name : str
                 The name of the queue to watch
                 
    counter_name : str
                   The name of the incrementing counter to watch. 
                   
    operation : str
                either 'insert' or 'update'. If 'insert', the sqlalchemy
                bulk_insert_mappings func is used to add new rows. If 'update', 
                the sqlalchemy bulks_update_mappings func is used.
    
    engine : obj
              A sqlalchemy engine. 
              
    stop_event : obj
                 A threading.Event object with set and is_set members. This is the
                 poison pill that can be set to terminate the thread.
    """
    # Use a pipe for efficient reads
    pipe = queue.pipe()
    while not stop_event.is_set():

        # Determine the read length of objects to pull from the cache
        # Cap the read length at 250 messages to improve memory usage and throughput
        read_length = int(queue.get(counter_name))
        if read_length > 250:
            read_length = 250
        # Pull the objects from the cache
        points = []
        measures = []
        
        # Pull the SRID dynamically from the model (database)
        rect_srid = Points.rectangular_srid
        lat_srid = Points.latitudinal_srid
        
        try:
            # Pop the messages off the queue
            for i in range(0, read_length):
                pipe.lpop(queue_name)
                pipe.decr(counter_name)
            msgs = pipe.execute()
        except:
            msgs = []
            time.sleep(5)

        for msg in msgs:
            msg = json.loads(msg, object_hook=object_hook)
            if isinstance(msg, dict):
                # A NULL id is not allowable, so pop if a NULL ID exists
                if msg['id'] == None:
                    msg.pop('id', None)

                # Since this avoids the ORM, need to map the table names manually
                msg['pointType'] = msg['pointtype']  
                adjusted = msg['adjusted']
            
                msg['adjusted'] = f'SRID={rect_srid};' + adjusted.wkt  # Geometries go in as EWKT
                msg['apriori'] = f'SRID={rect_srid};' + adjusted.wkt

                lon_og, lat_og, _ = reproject([adjusted.x, adjusted.y, adjusted.z],
                                    Points.semimajor_rad, Points.semiminor_rad,
                                    'geocent', 'latlon')
                lon, lat = og2oc(lon_og, lat_og, Points.semimajor_rad, Points.semiminor_rad)
                msg['geom'] = f'SRID={lat_srid};Point({lon} {lat})' 

                # Measures are removed and manually added later
                point_measures = msg.pop('measures', [])
                if point_measures:
                    measures.append(point_measures)
                
                points.append(msg)
            
        if not points:
            continue
        
        try:
            # Write the cached objects into the database 
            with engine.connect() as conn:
                resp = conn.execute(
                    insert(Points.__table__).returning(Points.__table__.c.id),points
                )
                pointids = [i[0] for i in resp.all()]

                # Measures are a list of lists. Associate each list with a pointid and then flatten the list
                for i, measure_set in enumerate(measures):
                    for measure in measure_set:
                        measure['pointid'] = pointids[i]
                        measure.pop('id', None)  # As above, remove the NULL id
                        # Remap field names because the ORM is NOT being used
                        measure['serialnumber'] = measure.pop('serial', None)
                        measure['measureType'] = measure.pop('measuretype')
                measures = [measure for sublist in measures for measure in sublist]
                conn.execute(
                    insert(Measures.__table__), measures)
        except:
            try:
                # Pop the messages off the queue
                for i in range(0, read_length):
                    pipe.rpush(queue_name, msgs[i])
                    pipe.incr(counter_name)
                msgs = pipe.execute()
            except:
                warnings.warn('Failed to push to DB and failed to repopulate queue.')
                time.sleep(5)

def watch_update_queue(queue, queue_name, counter_name, engine, stop_event, sleep_time=5):
    """
    A worker process to be launched in a thread that will asynchronously insert or update 
    objects in the Session using dicts pulled from a redis queue. Using this queuing approach
    many cluster jobs are able to push to the redis queue rapidly and then a single writer
    process can push the data back to the database.
    
    This function requires that the function called by the asynchronous cluster job INCR
    (increment) the counter_name key in the redis cache. This counter is INCR (incremented) 
    by cluster jobs to track how many messages have been pushed to the queue (queue_name). 
    This func then reads that many messages and DECR (de-increments) the counter by that
    many messages. This way this function only reads when data is present and reads can occur 
    asynchronously. This works becase the cluster job pushes to the right side of the redis
    list and this function reads n-messages from the left side.
    
    This method uses the sqlalchemy core interface for performance reasons. Therefore, some
    mundging of column names is used to ensure that the model to be processed matches the
    database column names.
    
    Parameters
    ----------
    queue : obj
            A Redis or StrictRedis connection instance
    
    queue_name : str
                 The name of the queue to watch
                 
    counter_name : str
                   The name of the incrementing counter to watch. 
                   
    operation : str
                either 'insert' or 'update'. If 'insert', the sqlalchemy
                bulk_insert_mappings func is used to add new rows. If 'update', 
                the sqlalchemy bulks_update_mappings func is used.
    
    engine : obj
              A sqlalchemy engine. 
              
    stop_event : obj
                 A threading.Event object with set and is_set members. This is the
                 poison pill that can be set to terminate the thread.
    """
    while not stop_event.is_set():
        # Determine the read length of objects to pull from the cache
        read_length = int(queue.get(counter_name))
        # Pull the objects from the cache
        measures = []
        
        for i in range(0, read_length):
            msg = json.loads(queue.lpop(queue_name), object_hook=object_hook)
            if isinstance(msg, dict):
                msg['_id'] = msg.pop('id', None)  # id is reserved by sqlalchemy on insert/update, remapped below
                measures.append(msg)

            # The message was successfully read, so atomically deincrement the counter
            queue.decr(counter_name)
                    
        # Write the updated measures to the db 
        if measures:
            with engine.connect() as conn:
                stmt = Measures.__table__.update().\
                            where(Measures.__table__.c.id == bindparam('_id')).\
                            values({'weight':bindparam('weight'),
                                    'measureIgnore':bindparam('ignore'),
                                    'templateMetric':bindparam('template_metric'),
                                    'templateShift':bindparam('template_shift'),
                                    'line': bindparam('line'),
                                    'sample':bindparam('sample'),
                                    'ChooserName':bindparam('choosername')})
                resp = conn.execute(
                    stmt, measures
                )

    time.sleep(sleep_time)



from collections import defaultdict
import json
import logging
from math import floor
import numbers
import sys
import time
import warnings
import logging

import numpy as np

from scipy.ndimage.measurements import center_of_mass
from skimage import transform as tf
from skimage import registration
from skimage import filters
from scipy import fftpack
from scipy.spatial import distance_matrix
import numpy as np
from sqlalchemy.sql.expression import bindparam

from matplotlib import pyplot as plt

from plio.io.io_gdal import GeoDataset

import pvl

import PIL
from PIL import Image

from autocnet.matcher.naive_template import pattern_match
from autocnet.matcher.mutual_information import mutual_information
from autocnet.spatial import isis
from autocnet.io.db.model import Measures, Points, Images, JsonEncoder
from autocnet.graph.node import NetworkNode
from autocnet.transformation import roi
from autocnet.transformation.affine import estimate_local_affine
from autocnet import spatial
from autocnet.utils.serializers import JsonEncoder

from sqlalchemy import inspect

PIL.Image.MAX_IMAGE_PIXELS = sys.float_info.max
log = logging.getLogger(__name__)


def check_image_size(imagesize):
    """
    Given an x,y tuple, ensure that the values
    are odd. Used by the subpixel template to also ensure
    that the template size is the one requested and not 2x
    the template size.

    Parameters
    ----------
    imagesize : tuple
                in the form (size_x, size_y)
    """
    if isinstance(imagesize, numbers.Number):
        imagesize = (int(imagesize), int(imagesize))


    x = imagesize[0]
    y = imagesize[1]

    if x % 2 == 0:
        x += 1
    if y % 2 == 0:
        y += 1
    x = floor(x/2)
    y = floor(y/2)
    return x,y

def subpixel_phase(reference_roi, moving_roi, affine=tf.AffineTransform(), **kwargs):
    """
    Apply the spectral domain matcher to a search and template image. To
    shift the images, the x_shift and y_shift, need to be subtracted from
    the center of the search image. It may also be necessary to apply the
    fractional pixel adjustment as well (if for example the center of the
    search is not an integer); this function do not manage shifting.

    Parameters
    ----------
    reference_roi : Object
                    An Roi object from autocnet, the reference image to use when computing subpixel offsets.
    moving_roi : Object
                  An Roi object from autocnet, the walking image to move around and make comparisons to
                  the reference roi.
    affine : Object
             Scikit image affine tranformation. This affine transformation is applied to the moving_roi
             before any comparisons are made

    Returns
    -------
    new_affine : Object
                 Scikit image affine transformation. An updated affine transformation that is the new
                 translation from the moving_roi to the reference_roi
    : tuple
      With the RMSE error and absolute difference in phase
    """
    reference_roi.clip()
    reference_image = reference_roi.clipped_array
    moving_roi.clip(affine=affine)
    walking_template = moving_roi.clipped_array

    if reference_image.shape != walking_template.shape:
        reference_size = reference_image.shape
        walking_size = walking_template.shape
        updated_size_x = int(min(walking_size[1], reference_size[1]))
        updated_size_y = int(min(walking_size[0], reference_size[0]))

        # Have to subtract 1 from even entries or else the round up that
        # occurs when the size is split over the midpoint causes the
        # size to be too large by 1.
        if updated_size_x % 2 == 0:
            updated_size_x -= 1
        if updated_size_y % 2 == 0:
            updated_size_y -= 1

        # Since the image is smaller than the requested size, set the size to
        # the current maximum image size and reduce from there on potential
        # future iterations.
        size = check_image_size((updated_size_x, updated_size_y))

        reference_roi.size_x, reference_roi.size_y = size
        moving_roi.size_x, moving_roi.size_y = size

        reference_image = reference_roi.clip()
        walking_template = moving_roi.clip(affine)
    (shift_y, shift_x), error, diffphase = registration.phase_cross_correlation(reference_image,
                                                                                walking_template,
                                                                                **kwargs)
    # get shifts in input pixel space 
    shift_x, shift_y = affine([shift_x, shift_y])[0]
    new_affine = tf.AffineTransform(translation=(-shift_x, -shift_y))
    return new_affine, error, diffphase

def subpixel_template(reference_roi, 
                      moving_roi, 
                      affine=tf.AffineTransform(), 
                      func=pattern_match,
                      **kwargs):
    """
    Uses a pattern-matcher on subsets of two images determined from the passed-in keypoints and optional sizes to
    compute an x and y offset from the search keypoint to the template keypoint and an associated strength.
    Parameters
    ----------
    reference_roi : autocnet.roi.Roi
                    Roi object representing the reference image 
    
    moving_roi : autocnet.roi.Roi
                 Roi object representing the moving image, this image is registered to the reference_roi   
    
    affine : skimage.transform.AffineTransform
             scikit-image Affine transformation, used as a seed transform that 

    func : callable 
           Some subpixel template matching function. Default autocnet.matcher.naive_template.pattern_match 
    
    kwargs : dict 
             keyword args for func 

    Returns
    -------
    affine : skimage.transform.AffineTransform
             Affine transform containing new translations to be applied to the x and y attributes of 
             the input moving_roi. For example, returned_affine((moving_roi.x, moving_roi.y))
    
    strength : float
               Strength of the correspondence in the range [-1, 1]
    
    corrmap : np.array 
              containing correlation coefficients computed at each tested location in the reference_roi
    
    See Also
    --------
    autocnet.matcher.naive_template.pattern_match : for the kwargs that can be passed to the matcher
    autocnet.matcher.naive_template.pattern_match_autoreg : for the kwargs that can be passed to the autoreg style matcher
    """

    try:
        s_image_dtype = isis.isis2np_types[pvl.load(reference_roi.data.file_name)["IsisCube"]["Core"]["Pixels"]["Type"]]
    except:
        s_image_dtype = None

    try:
        d_template_dtype = isis.isis2np_types[pvl.load(moving_roi.data.file_name)["IsisCube"]["Core"]["Pixels"]["Type"]]
    except:
        d_template_dtype = None
    
    
    # In ISIS, the reference image is the search and moving image is the pattern.
    reference_roi.clip(dtype = s_image_dtype)
    ref_clip = reference_roi.clipped_array
    moving_roi.clip(affine=affine, dtype=d_template_dtype)
    moving_clip = moving_roi.clipped_array
    
    if moving_clip.var() == 0:
        warnings.warn('Input ROI has no variance.')
        return [None] * 3

    if (ref_clip is None) or (moving_clip is None):
        return None, None, None

    matcher_shift_x, matcher_shift_y, metrics, corrmap = func(moving_clip, ref_clip, **kwargs)
    if matcher_shift_x is None:
        return None, None, None

    # shift_x, shift_y are in the affine transformed space. They are relative to center of the affine transformed ROI.
    # Apply the shift to the center of the moving roi to the center of the reference ROI in index space.
    new_center_x = moving_roi.clip_center[0] + matcher_shift_x
    new_center_y = moving_roi.clip_center[1] + matcher_shift_y
    
    new_x, new_y = moving_roi.clip_coordinate_to_image_coordinate(new_center_x, new_center_y)
    
    new_affine = tf.AffineTransform(translation=(-(moving_roi.x - new_x),
                                                 -(moving_roi.y - new_y)))
    
    return new_affine, metrics, corrmap


def iterative_phase(reference_roi, moving_roi, affine=tf.AffineTransform(), reduction=11, convergence_threshold=0.1, max_dist=50, **kwargs):
    """
    Iteratively apply a subpixel phase matcher to source (s_img) and destination (d_img)
    images. The size parameter is used to set the initial search space. The algorithm
    is recursively applied to reduce the total search space by reduction until the convergence criteria
    are met. Convergence is defined as the point at which the computed shifts (x_shift,y_shift) are
    less than the convergence_threshold. In instances where the size is reducted to 1 pixel the
    algorithm terminates and returns None.

    Parameters
    ----------
    reference_roi : Object
                    An Roi object from autocnet, the reference image to use when computing subpixel offsets.
                    Contains either an ndarray or a GeoDataset Object
    moving_roi : Object
                  An Roi object from autocnet, the walking image to move around and make comparisons to
                  the reference roi. Contains either an ndarray or a GeoDataset Object
    size : tuple
           Size of the template in the form (x,y)
    reduction : int
                With each recursive call to this func, the size is reduced by this amount
    convergence_threshold : float
                            The value under which the result can shift in the x and y directions to force a break

    Returns
    -------
    metrics : tuple
              A tuple of metrics. In the case of the phase matcher this are difference
              and RMSE in the phase dimension.

    See Also
    --------
    subpixel_phase : the function that applies a single iteration of the phase matcher
    """
    # get initial destination location
    dsample, dline = moving_roi.x, moving_roi.y
    dx, dy = moving_roi.x, moving_roi.y

    size = (moving_roi.size_x, moving_roi.size_y)
    original_size = (moving_roi.size_x, moving_roi.size_y)

    subpixel_affine = tf.AffineTransform()

    while True:
        new_subpixel_affine, error, diffphase = subpixel_phase(reference_roi, moving_roi, affine, **kwargs)
        # Compute the amount of move the matcher introduced
        delta_dx, delta_dy = abs(new_subpixel_affine.translation)
        subpixel_affine += new_subpixel_affine
        dx, dy = subpixel_affine.inverse((reference_roi.x, reference_roi.y))[0]
        moving_roi.x, moving_roi.y = dx, dy

        # Break if the solution has converged
        size = (size[0] - reduction, size[1] - reduction)
        moving_roi.size_x, moving_roi.size_y = size
        reference_roi.size_x, reference_roi.size_y = size
        dist = np.linalg.norm([dsample-dx, dline-dy])

        if min(size) < 1:
            return None, None, None
        if delta_dx <= convergence_threshold and\
           delta_dy <= convergence_threshold and\
           abs(dist) <= max_dist:
            break

    moving_roi.size_x, moving_roi.size_y = original_size
    reference_roi.size_x, reference_roi.size_y = original_size
    moving_roi.x, moving_roi.y = dsample, dline
    return subpixel_affine, error, diffphase

def subpixel_register_point(pointid,
                            cost_func=lambda x,y: 1/x**2 * y,
                            threshold=0.005,
                            ncg=None,
                            match_func=subpixel_template,
                            match_kwargs={},
                            use_cache=False,
                            verbose=False,
                            chooser='subpixel_register_point',
                            **kwargs):

    """
    Given some point, subpixel register all of the measures in the point to the
    first measure.

    Parameters
    ----------
    pointid : int or obj
              The identifier of the point in the DB or a Points object

    cost_func : func
                A generic cost function accepting two arguments (x,y), where x is the
                distance that a point has shifted from the original, sensor identified
                intersection, and y is the correlation coefficient coming out of the
                template matcher.

    threshold : numeric
                measures with a cost <= the threshold are marked as ignore=True in
                the database.
    ncg : obj
          the network candidate graph that the point is associated with; used for
          the DB session that is able to access the point.

    geom_func : callable
                function used to tranform the source and/or destination image before
                running a matcher.

    match_func : callable
                 subpixel matching function to use registering measures

    use_cache : bool
                If False (default) this func opens a database session and writes points
                and measures directly to the respective tables. If True, this method writes
                messages to the point_insert (defined in ncg.config) redis queue for
                asynchronous (higher performance) inserts.
    """

    if not ncg.Session:
        raise BrokenPipeError('This func requires a database session from a NetworkCandidateGraph.')

    if isinstance(pointid, Points):
        pointid = pointid.id

    t1 = time.time()
    with ncg.session_scope() as session:
        measures = session.query(Measures).filter(Measures.pointid == pointid).order_by(Measures.id).all()
        point = session.query(Points).filter(Points.id == pointid).one()
        reference_index = point.reference_index
        t2 = time.time()
        log.info(f'Query took {t2-t1} seconds to find the measures and reference measure.')
        # Get the reference measure. Previously this was index 0, but now it is a database tracked attribute
        source = measures[reference_index]

        log.info(f'Using measure {source.id} on image {source.imageid}/{source.serial} as the reference.')
        log.info(f'Measure reference index is: {reference_index}')
        source.template_metric = 1
        source.template_shift = 0
        source.phase_error = 0
        source.phase_diff = 0
        source.phase_shift = 0

        sourceid = source.imageid
        sourceres = session.query(Images).filter(Images.id == sourceid).one()
        source_node = NetworkNode(node_id=sourceid, image_path=sourceres.path)
        source_node.parent = ncg
        t3 = time.time()
        log.info(f'Query for the image to use as source took {t3-t2} seconds.')
        log.info(f'Attempting to subpixel register {len(measures)-1} measures for point {pointid}')
        nodes = {}
        for measure in measures:
            res = session.query(Images).filter(Images.id == measure.imageid).one()
            nodes[measure.imageid] = NetworkNode(node_id=measure.imageid, image_path=res.path)
        session.expunge_all()

    resultlog = []
    updated_measures = []
    for i, measure in enumerate(measures):
        if i == reference_index:
            continue

        currentlog = {'measureid':measure.id,
                    'status':''}
        cost = None

        destination_node = nodes[measure.imageid]

        reference_roi = roi.Roi(source_node.geodata, 
                                source.apriorisample, 
                                source.aprioriline,
                                size_x=match_kwargs['image_size'][0],
                                size_y=match_kwargs['image_size'][1])
        moving_roi = roi.Roi(destination_node.geodata, 
                             measure.apriorisample, 
                             measure.aprioriline,
                             size_x=match_kwargs['template_size'][0],
                             size_y=match_kwargs['template_size'][1],
                             buffer=5)

        baseline_affine = estimate_local_affine(reference_roi, moving_roi)

        # Updated so that the affine used is computed a single time.
        # Has not scale or shear or rotation.
        updated_affine, maxcorr, _ = subpixel_template(reference_roi,
                                                        moving_roi,
                                                        affine=baseline_affine)
        
        if updated_affine is None:
            log.warn('Unable to match with this parameter set.')
            currentlog['status'] = f"subpixel registration failed on measure {measure.id}"
            resultlog.append(currentlog)
            if measure.weight is None:
                measure.ignore = True # Geom match failed and no previous sucesses
            updated_measures.append(measure)
            continue
        new_x, new_y = updated_affine([measure.apriorisample, measure.aprioriline])[0]

        if new_x == None or new_y == None:
            currentlog['status'] = f'Failed to register measure {measure.id}.'
            resultlog.append(currentlog)
            if measure.weight is None:
                measure.ignore = True # Unable to geom match and no previous sucesses
            updated_measures.append(measure)
            continue

        measure.template_metric = maxcorr
        dist = np.linalg.norm([measure.apriorisample-new_x, measure.aprioriline-new_y])
        measure.template_shift = dist

        cost = cost_func(measure.template_shift, measure.template_metric)

        log.info(f'Current Cost: {cost},  Current Weight: {measure.weight}')

        # Check to see if the cost function requirement has been met
        if measure.weight and cost <= measure.weight:
            currentlog['status'] = f'Previous match provided better correlation. {measure.weight} > {cost}.'
            resultlog.append(currentlog)
            updated_measures.append(measure)
            continue

        if cost <= threshold:
            currentlog['status'] = f'Cost failed. Distance calculated: {measure.template_shift}. Metric calculated: {measure.template_metric}.'
            resultlog.append(currentlog)
            updated_measures.append(measure)
            if measure.weight is None:
                measure.ignore = True # Threshold criteria not met and no previous sucesses
            continue

        # Update the measure
        measure.sample = new_x
        measure.line = new_y
        measure.weight = cost
        measure.choosername = chooser

        # In case this is a second run, set the ignore to False if this
        # measures passed. Also, set the source measure back to ignore=False
        measure.ignore = False
        # Maybe source?
        source.ignore = False
        updated_measures.append(measure)
        currentlog['status'] = f'Success. Distance shifted: {measure.template_shift}. Metric: {measure.template_metric}.'
        resultlog.append(currentlog)

    # Once here, update the source measure (possibly back to ignore=False)
    updated_measures.append(source)

    if use_cache:
        ncg.redis_queue.rpush(ncg.measure_update_queue,
                              *[json.dumps(measure.to_dict(_hide=[]), cls=JsonEncoder) for measure in updated_measures])
        ncg.redis_queue.incr(ncg.measure_update_counter, amount=len(updated_measures))

    else:
        # Commit the updates back into the DB
        with ncg.session_scope() as session:
            for m in updated_measures:
                ins = inspect(m)
                session.add(m)

    return resultlog

def subpixel_register_points(subpixel_template_kwargs={'image_size':(251,251)},
                             cost_kwargs={},
                             threshold=0.005,
                             Session=None):
    """
    Serial subpixel registration of all of the points in a given DB table.

    Parameters
    ----------
    Session : obj
              A SQLAlchemy Session factory.

    pointid : int
              The identifier of the point in the DB

    subpixel_template_kwargs : dict
                               Ay keyword arguments passed to the template matcher

    cost : func
           A generic cost function accepting two arguments (x,y), where x is the
           distance that a point has shifted from the original, sensor identified
           intersection, and y is the correlation coefficient coming out of the
           template matcher.

    threshold : numeric
                measures with a cost <= the threshold are marked as ignore=True in
                the database.
    """
    if not Session:
        raise BrokenPipeError('This func requires a database session.')
    session = Session()
    pointids = [point.id for point in session.query(Points)]
    session.close()
    for pointid in pointids:
        subpixel_register_point(pointid,
                                subpixel_template_kwargs=subpixel_template_kwargs,
                                **cost_kwargs)

def register_to_base(pointid,
                     base_image,
                     cost_func=lambda x, y: y == np.max(x),
                     ncg=None,
                     geom_func='simple',
                     geom_kwargs={"size_x": 16, "size_y": 16},
                     match_func='classic',
                     match_kwargs={},
                     verbose=False,
                     **kwargs):
    """
    """

    if not ncg.Session:
     raise BrokenPipeError('This func requires a database session from a NetworkCandidateGraph.')

    geom_func = check_geom_func(geom_func)
    match_func = check_match_func(match_func)
    session = ncg.Session()

    if isinstance(base_image, str):
        base_image = GeoDataset(base_image)

    if isinstance(pointid, Points):
        point = pointid
        pointid = pointid.id

    with ncg.session_scope() as session:
        if isinstance(pointid, Points):
            point = pointid
            pointid = point.id
        else:
            point = session.query(Points).filter(Points.id == pointid).one()

        # Get all of the measures associated with the given point
        measures = point.measures

        # Attempt to project the point into the base image
        bpoint = spatial.isis.point_info(base_image.file_name, point.geom.x, point.geom.y, 'ground')
        if bpoint is None:
            log.warning('unable to find point in ground image')
            # Need to set the point to False
            return
        bline = bpoint.get('Line')
        bsample = bpoint.get('Sample')

        # Setup a cache so that we can get the file handles one time instead of
        # once per measure in the measures list.
        image_cache = {}

        # list of matching results in the format:
        # [measure_id, measure_index, x_offset, y_offset, offset_magnitude]
        match_results = []

        # Step over all of the measures (images) that are not the base image
        for measure_index, measure in enumerate(measures):
            res = session.query(Images).filter(Images.id == measure.imageid).one()
            try:
                measure_image = image_cache[res.id]
            except:
                measure_image = GeoDataset(measure_image.path)
                image_cache[res.id] = measure_image

            # Attempt to match the base
            try:
                log.info(f'prop point: base_image: {base_image}')
                log.info(f'prop point: dest_image: {measure_image}')
                log.info(f'prop point: (sx, sy): ({measure.sample}, {measure.line})')
                x, y, dist, metrics = geom_func(base_image, measure_image,
                        bsample, bline,
                        match_func = match_func,
                        match_kwargs = match_kwargs,
                        verbose=verbose,
                        **geom_kwargs)

            except Exception as e:
                raise Exception(e)
                match_results.append(e)
                continue

            match_results.append([measure.id, measure_index, x, y,
                                 metrics, dist, base_image.file_name, measure_image.file_name])

    if verbose:
      log.info("Match Results: ", match_results)

    # Clean out any instances where None has been return by the geom matcher.
    match_results = np.copy(np.array([res for res in match_results if isinstance(res, list) and all(r is not None for r in res)]))

    # If all of the results are None, this point was not matchable
    if match_results.shape[0] == 0:
        raise Exception("Point with id {pointid} has no measure that matches criteria, reference measure will remain unchanged")

    # Compute the cost function for each match using the
    costs = [cost_func(match_results[:,3], match[3]) for match in match_results]

    if verbose:
      log.info("Values:", costs)

    # column index 3 is the metric returned by the geom matcher
    best_results = match_results[np.argmax(costs)]

    if verbose:
        log.info("match_results final length: ", len(match_results))
        log.info("best_results length: ", len(best_results))
        log.info("Full results: ", best_results)
        log.info("Winning CORRs: ", best_results[3], "Base Pixel shifts: ", best_results[4])
        log.info('\n')

    if len(best_results[3])==1 or best_results[3] is None:
        raise Exception("Point with id {pointid} has no measure that matches criteria, reference measure will remain unchanged")

    # Finally, update the point that will be the reference
    with ncg.session_scope() as session:
       measure = session.query(Measures).filter(Measures.id == best_results[0]).one()
       measure.sample = best_results[2]
       measure.line = best_results[3]

       point = session.query(Points).filter(Points.id == pointid).one()
       point.ref_measure = best_results[1]
    return

def estimate_logpolar_transform(img1, img2, low_sigma=0.5, high_sigma=30, verbose=False):
    """
    Estimates the rotation and scale difference for img1 that maps to img2 using phase cross correlation on a logscale projection.

    Scale and angular changes in cartesian space become translation in log-polar space. Translation from subpixel registration
    in log-polar space can then be translated into scale/rotation change in the original cartesian images. This scale + roation
    change estimation is then returned as an affine transform object. This can then be used before other subpixel registration
    methods to enable scale+rotation invariance.

    See Also
    --------

    skimage.filters.difference_of_gaussians : Bandpass filtering using a difference of gaussians
    skimage.filters.window : Simple wondowing function to remove spectral leakage along the axes in the fourier transform

    References
    ----------

    .. [1] Rittavee Matungka. 2009. Studies on log-polar transform for image registration and improvements
       using adaptive sampling and logarithmic spiral. Ph.D. Dissertation. Ohio State University, USA. Advisor(s) Yuan F. Zheng.
       Order Number: AAI3376091.
    .. [2] https://github.com/polakluk/fourier-mellin
    .. [3] https://scikit-image.org/docs/stable/auto_examples/registration/plot_register_rotation.html

    Parameters
    ----------

    img1: np.ndarray
          The source image, this is the image that is used as a base as img2 is registered to the center on img1

    img2: np.ndarray
          The image that will be moved to match img1

    low_sigma : float, list, np.array
                The low standard deviation for the Gaussian kernel used in the difference of gaussians filter. This reccomended
                to remove high frequency noise from the image before the log-polar projection as high frequency noise negatively impact registration
                in log-polar space. The lower the sigma, the sharper the resulting image will be. Use a small low_sigma with a large high_sigma
                to remove high frequency noise. Default is 0.5.

    high_sigma : float, list, np.array
                Standard deviation for the Gaussian kernel with the larger sigmas across all axes used in the difference of gaussians filter. This reccomended
                to remove high frequency noise from the image before the log-polar projection as high frequency noise negatively impact registration
                in log-polar space. The higher this sigma compared to the low_sigma, the more detail will be preserved. Use a small low_sigma with a large high_sigma
                to remove high frequency noise. A high sigma equal to ~1.6x the low sigma is reccomended for edge detection, so consider high_sigmas >= low_sigma*1.6. Default is 30.

    verbose : bool
              If true, prints out information detailing the registration process

    Returns
    -------
    : skimage.transform.SimilarityTansform
      Scikit-image affine transformation object containing rotation and scale information to warp img1 to img2

    """
    # First, band-pass filter both images
    img1 = filters.difference_of_gaussians(img1, low_sigma, high_sigma)
    img2 = filters.difference_of_gaussians(img2, low_sigma, high_sigma)

    # window images
    wimg1 = img1 * (filters.window('hann', img1.shape))
    wimg2 = img2 * (filters.window('hann', img2.shape))

    # work with shifted FFT magnitudes
    img1_fs = np.abs(fftpack.fftshift(fftpack.fft2(wimg1)))
    img2_fs = np.abs(fftpack.fftshift(fftpack.fft2(wimg2)))

    # Create log-polar transformed FFT mag images and register
    shape = img1_fs.shape
    radius = shape[0] // 4  # only take lower frequencies
    warped_img1_fs = tf.warp_polar(img1_fs, radius=radius, output_shape=shape,
                                 scaling='log', order=0)
    warped_img2_fs = tf.warp_polar(img2_fs, radius=radius, output_shape=shape,
                               scaling='log', order=0)

    warped_img1_fs = warped_img1_fs[:shape[0] // 2, :]
    warped_img2_fs = warped_img2_fs[:shape[0] // 2, :]
    shifts, error, phasediff = registration.phase_cross_correlation(warped_img1_fs,
                                                       warped_img2_fs,
                                                       upsample_factor=10)

    # Use translation parameters to calculate rotation and scaling parameters
    shiftr, shiftc = shifts[:2]
    recovered_angle = -(360 / shape[0]) * shiftr
    klog = shape[1] / np.log(radius)
    shift_scale = np.exp(shiftc / klog)
    
    if recovered_angle < - 45.0:
        recovered_angle += 180
    else:
        if recovered_angle > 90.0:
            recovered_angle -= 180

    if verbose:
        fig, axes = plt.subplots(2, 2, figsize=(8, 8))
        ax = axes.ravel()
        ax[0].set_title("Original Image FFT\n(magnitude; zoomed)")
        center = np.array(shape) // 2
        ax[0].imshow(img1_fs[center[0] - radius:center[0] + radius,
                              center[1] - radius:center[1] + radius],
                     cmap='magma')
        ax[1].set_title("Modified Image FFT\n(magnitude; zoomed)")
        ax[1].imshow(img2_fs[center[0] - radius:center[0] + radius,
                            center[1] - radius:center[1] + radius],
                     cmap='magma')
        ax[2].set_title("Log-Polar-Transformed\nOriginal FFT")
        ax[2].imshow(warped_img1_fs, cmap='magma')
        ax[3].set_title("Log-Polar-Transformed\nModified FFT")
        ax[3].imshow(warped_img2_fs, cmap='magma')
        fig.suptitle('Working in frequency domain can recover rotation and scaling')
        plt.show()

        log.info(f"Recovered value for cc rotation: {recovered_angle}")
        log.info(f"Recovered value for scaling difference: {shift_scale}")

    # offset by the center of the image, scikit's ceter image rotation is defined by `axis / 2 - 0.5`
    shift_y, shift_x = np.asarray(img1.shape) / 2 - 0.5
    tf_scale = tf.SimilarityTransform(scale=shift_scale)
    tf_rotate = tf.SimilarityTransform(rotation=np.deg2rad(recovered_angle))
    tf_shift = tf.SimilarityTransform(translation=[-shift_x, -shift_y])
    tf_shift_inv = tf.SimilarityTransform(translation=[shift_x, shift_y])

    tf_rotate_from_center = (tf_shift + (tf_rotate + tf_shift_inv))
    return tf.SimilarityTransform((tf_rotate_from_center + tf_scale)._inv_matrix)


def fourier_mellen(ref_image, moving_image, affine=tf.AffineTransform(), verbose=False, phase_kwargs={}):
    """
    Iterative phase registration using a log-polar projection to estimate an affine for scale and roation invariance.


    Parameters
    ----------

    ref_image: np.ndarray
               The source image, this is the image that is used as a base as img2 is registered to the center on img1

    moving_image: np.ndarray
                  The image that will be moved to match img1

    verbose : bool
              If true, prints out information detailing the registration process

    phase_kwargs : dict
                   Parameters to be passed into the iterative_phase matcher

    Returns
    -------

    : float
      The new x coordinate for moving_image that registers to the center of center_image

    : float
      The new y coordinate for moving_image that registers to the center of center_image

    : float
      Error returned by the iterative phase matcher
    """
    # Get the affine transformation for scale + rotation invariance
    affine = estimate_logpolar_transform(ref_image.array, moving_image.array, verbose=verbose)

    # warp the source image to match the destination
    ref_warped = ref_image.clip(affine)
    sx, sy = affine.inverse(np.asarray(ref_image.array.shape)/2)[0]

    # get translation with iterative phase
    newx, newy, error = iterative_phase(sx, sy, sx, sy, ref_warped, moving_image, **phase_kwargs)

    if verbose:
        fig, axes = plt.subplots(2, 2, figsize=(8, 8))
        ax = axes.ravel()

        ax[0].imshow(ref_warped)
        ax[0].set_title("Image 1 Transformed")
        ax[0].axhline(y=sy, color="red", linestyle="-", alpha=1, linewidth=1)
        ax[0].axvline(x=sx, color="red", linestyle="-", alpha=1, linewidth=1)

        ax[2].imshow(ref_image)
        ax[2].set_title("Image 1")
        ax[2].axhline(y=ref_image.array.shape[0]/2, color="red", linestyle="-", alpha=1, linewidth=1)
        ax[2].axvline(x=ref_image.array.shape[1]/2, color="red", linestyle="-", alpha=1, linewidth=1)

        ax[1].imshow(moving_image)
        ax[3].imshow(moving_image)

        if not newx or not newy:
            ax[1].set_title("Image 2 REGISTRATION FAILED")
            ax[3].set_title("Image 2 REGISTRATION FAILED")
        else :
            ax[3].set_title("Image 2 Registered")
            ax[1].axhline(y=newy, color="red", linestyle="-", alpha=1, linewidth=1)
            ax[1].axvline(x=newx, color="red", linestyle="-", alpha=1, linewidth=1)
            ax[3].axhline(y=newy, color="red", linestyle="-", alpha=1, linewidth=1)
            ax[3].axvline(x=newx, color="red", linestyle="-", alpha=1, linewidth=1)

    return newx, newy, error


def subpixel_register_point_smart(pointid,
                            cost_func=lambda x,y: 1/x**2 * y,
                            ncg=None,
                            parameters=[],
                            chooser='subpixel_register_point_smart',
                            verbose=False):

    """
    Given some point, subpixel register all of the measures in the point to the
    reference measure.

    Parameters
    ----------
    
    pointid : int or obj
              The identifier of the point in the DB or a Points object

    cost_func : func
                A generic cost function accepting two arguments (x,y), where x is the
                distance that a point has shifted from the original, sensor identified
                intersection, and y is the correlation coefficient coming out of the
                template matcher.

    ncg : obj
          the network candidate graph that the point is associated with; used for
          the DB session that is able to access the point.

    parameters : list
                 of dicts containing "match_kwargs" used for specified match_func.
                 The passed parameters describe image and template chips that are tested.
                 For example parameters = [
                 {'match_kwargs': {'image_size':(121,121), 'template_size':(61,61)}},
                 {'match_kwargs': {'image_size':(151,151), 'template_size':(67,67)}},
                 {'match_kwargs': {'image_size':(181,181), 'template_size':(73,73)}}]
    """


    if not ncg.Session:
        raise BrokenPipeError('This func requires a database session from a NetworkCandidateGraph.')

    if isinstance(pointid, Points):
        pointid = pointid.id

    with ncg.session_scope() as session:
        # Order by is important here because the measures get ids in sequential order when initially placed
        # and the reference_index is positionally linked to the ordered vector of measures.
        measures = session.query(Measures).filter(Measures.pointid == pointid).order_by(Measures.id).all()
        point = session.query(Points).filter(Points.id == pointid).one()
        reference_index = point.reference_index

        # Get the reference measure to instantiate the source node. All other measures will
        # match to the source node.
        source = measures[reference_index]
        reference_index_id = source.imageid

        log.info(f'Using measure {source.id} on image {source.imageid}/{source.serial} as the reference.')
        log.info(f'Measure reference index is: {reference_index}')

        # Build a node cache so that this is an encapsulated database call. Then nodes
        # can be pulled from the lookup sans database.
        nodes = {}
        for measure in measures:
            res = session.query(Images).filter(Images.id == measure.imageid).one()
            nn = NetworkNode(node_id=measure.imageid, image_path=res.path)
            nn.parent = ncg
            nodes[measure.imageid] = nn

        session.expunge_all()

    log.info(f'Attempting to subpixel register {len(measures)-1} measures for point {pointid}')
    # Set the reference image
    source_node = nodes[reference_index_id]

    log.info(f'Source: sample: {source.sample} | line: {source.line}')
    updated_measures = []
    for i, measure in enumerate(measures):

        # If this is the reference node, do not attempt to match it.
        if i == reference_index:
            continue

        cost = None
        destination_node = nodes[measure.imageid]
        log.info(f'Registering measure {measure.id} (image: {measure.imageid})')

        # Compute the baseline metrics using the smallest window
        size_x = np.inf
        size_y = np.inf
        for p in parameters:
            match_kwarg = p['match_kwargs']
            if match_kwarg['template_size'][0] < size_x:
                size_x = match_kwarg['template_size'][0]
            if match_kwarg['template_size'][1] < size_y:
                size_y = match_kwarg['template_size'][1]

        reference_roi = roi.Roi(source_node.geodata, 
                                source.apriorisample, 
                                source.aprioriline, 
                                size_x=size_x, 
                                size_y=size_y, 
                                buffer=0)
        moving_roi = roi.Roi(destination_node.geodata, 
                             measure.apriorisample, 
                             measure.aprioriline, 
                             size_x=size_x, 
                             size_y=size_y, 
                             buffer=20)

        try:
            baseline_affine = estimate_local_affine(reference_roi,
                                                    moving_roi)
        except Exception as e:
            log.exception(e)
            m = {'id': measure.id,
                 'sample':measure.apriorisample,
                 'line':measure.aprioriline,
                 'status':False,
                 'choosername':chooser}
            updated_measures.append([None, None, m])
            continue

        reference_roi.clip()
        reference_clip = reference_roi.clipped_array

        # If the image read center plus buffer is outside the image, clipping
        # raises an index error. Handle and set the measure false.
        try:
            moving_roi.clip(affine=baseline_affine)
            moving_clip = moving_roi.clipped_array
        except Exception as e:
            log.error(e)
            m = {'id': measure.id,
                 'sample':measure.apriorisample,
                 'line':measure.aprioriline,
                 'status':False,
                 'choosername':chooser}
            updated_measures.append([None, None, m])
            continue

        reference_nan = np.isnan(reference_clip).any()
        moving_nan = np.isnan(moving_clip).any()
        if reference_nan or moving_nan:
            if reference_nan:
                log.warning(f'Unable to process due to NaN values in the reference data.')
            if moving_nan:
                log.warning(f'Unable to process due to NaN values in the moving data.')
            m = {'id': measure.id,
                    'status': False,
                    'choosername': chooser}
            updated_measures.append([None, None, m])
            continue

        if reference_clip.shape != moving_clip.shape:
            log.warning('Unable to process. ROIs are different sizes for MI matcher')
            m = {'id': measure.id,
                 'status': False,
                 'choosername': chooser}
            updated_measures.append([None, None, m])
            continue

        # Compute the a priori template and MI correlations with no shifts allowed. 
        _, baseline_corr, _ = subpixel_template(reference_roi, 
                                                moving_roi,
                                                affine=baseline_affine)
        
        baseline_mi = 0 #mutual_information_match(reference_roi, 
                        #                         moving_roi, 
                        #                         affine=baseline_affine)
        
        log.info(f'Baseline MI: {baseline_mi} | Baseline Corr: {baseline_corr}')
        
        for parameter in parameters:
            match_kwargs = parameter['match_kwargs']

            reference_roi = roi.Roi(source_node.geodata,
                                    source.apriorisample,
                                    source.aprioriline,
                                    size_x=match_kwargs['image_size'][0],
                                    size_y=match_kwargs['image_size'][1], 
                                    buffer=0)
            moving_roi = roi.Roi(destination_node.geodata,
                                 measure.apriorisample,
                                 measure.aprioriline,
                                 size_x=match_kwargs['template_size'][0],
                                 size_y=match_kwargs['template_size'][1], 
                                 buffer=20)
            if verbose:
                fig, axes = plt.subplots(1,2)
                reference_roi.clip()
                moving_roi.clip(affine=baseline_affine)
                axes[0].imshow(reference_roi.clipped_array, cmap='Greys')
                axes[1].imshow(moving_roi.clipped_array, cmap='Greys')
                plt.show()
            try:  # Handle the case where the parameter set plus the buffer is outside
                  # the image and the clip raises an index error.
                updated_affine, maxcorr, _ = subpixel_template(reference_roi,
                                                            moving_roi,
                                                            affine=baseline_affine)
            except Exception as e:
                log.error(e)
                updated_affine = None

            if updated_affine is None:
                log.warning(f'Unable to match with this parameter set.')
                continue
            else:
                mi_metric=0
                metric = maxcorr
                new_x, new_y = updated_affine([measure.apriorisample, measure.aprioriline])[0]
                
                dist = np.linalg.norm([measure.aprioriline-new_x, 
                                      measure.apriorisample-new_y])
                cost = cost_func(dist, metric)

                m = {'id': measure.id,
                    'sample':new_x,
                    'line':new_y,
                    'weight':cost,
                    'choosername':chooser,
                    'template_metric':metric,
                    'template_shift':dist,
                    'mi_metric': mi_metric,
                    'status': True}
                log.info(f'METRIC: {metric}| SAMPLE: {new_x} | LINE: {new_y} | MI: {mi_metric}')
            
            updated_measures.append([baseline_mi, baseline_corr, m])

    # Baseline MI, Baseline Correlation, updated measures to select from
    return updated_measures

def check_for_shift_consensus(shifts, tol=0.1):
    """
    Find matched locations from a set of multiple different solutions that have
    the same position within some user supplied tolerance. If the distance between
    two measures (shifts) is <= the tolerance, the measures are considered to have
    found consensus.

    This doc string uses 'measure' to describe each solution found by a subpixel
    matching attempt. If n-attempts are made, using n-different parameter sets,
    this function will find shift consensus between those n-different solutions.

    The function works by computing the full distance matrix between all solutions,
    generating a boolean mask for distances less than the tolerance, then generating
    a vector of column sums where the sum is the number of inliers, and finally,
    returning a boolean vector where the column sums are greater than 2.

    Parameters
    ----------
    shifts : ndarray
             (n,2) array of (x,y) coordinates representing the subpixel registered
             measure locations. n must be >= 3.

    tol : float
          The tolerance value required for measures to be inliers. Distances between
          points less than or equal to the tolerance are inliers. In pixel space.

    Returns
    -------
     : ndarray
       (n,1) boolean array where the nth element corresponds to the nth measure
       in the shifts input array. True values indicate that the measure has shift
       consensus with at least 2 other measures
    """
    dists = distance_matrix(shifts, shifts)
    inliers = dists <= tol
    col_sums = np.sum(inliers, 1)
    # The distance matrix is zero diagonal, so 2+ means one other matcher found
    # a close location
    return col_sums >= 2
    
def decider(measures, tol=0.5):
    """
    The logical decision function that determines which measures would be updated
    with subpixel registration or ignored. The function iterates over the measures,
    looks for shift consensus between subpixel registration runs.

    Parameters
    ----------
    measures : list
               A list of candidate measures (dicts) objects from the smart subpixel matcher

    tol : float
          The tolerance value required for points to be inliers. Distances between
          points less than or equal to the tolerance are inliers. In pixel space.

    Returns
    -------
    measures_to_update : list
                         of measures (dicts) to be updated to subpixel accuracy

    measures_to_set_false : list
                            of meaure ids to be ignored beause theu fail the consensus
                            building approach
    """
    by_id = defaultdict(list)
    measures_to_set_false = []
    for m in measures:
        baseline_mi = m[0]
        baseline_corr = m[1]
        m = m[2]
        if m['status'] and m['mi_metric'] is not None and m['template_metric'] is not None:
            choosername = m['choosername']
            by_id[m['id']].append([m['line'],
                                  m['sample'],
                                  m['mi_metric'],
                                  m['template_metric'],
                                  baseline_mi,
                                  baseline_corr,
                                  m['template_shift']])
        else:
            measures_to_set_false.append(m['id'])

    measures_to_update = []
    for k, v in by_id.items():
        v = np.asarray(v)
        mi = v[:,2]
        corr = v[:,3]
        baseline_mi = v[:,4]
        baseline_corr = v[:,5]
        cost = (baseline_mi - mi) + (baseline_corr - corr)

        # At least two of the correlators need to have found a soln within 0.5 pixels.
        shift_mask = check_for_shift_consensus(v[:,:2], tol=tol)

        # This is formulated as a minimization, so the best is the min cost
        best_cost = np.argmin(cost)

        if shift_mask[best_cost] == False:
            # The best cost does not have positional consensus
            measures_to_set_false.append(k)
        else:
            best_measure = v[best_cost]
            m = {'id':k,
                 'line': best_measure[0],
                 'sample': best_measure[1],
                 'weight': cost[best_cost],
                 'template_metric': best_measure[3],
                 'template_shift': best_measure[6],
                 'choosername': choosername,
                 'ignore':False,
                 'best_parameter_index': best_cost}
            measures_to_update.append(m)
    # A measure could have one bad regitration and get set false, if a different parameter set passed,
    # remove from the set false list.
    ids_to_update = [d['id'] for d in measures_to_update]
    measures_to_set_false = [i for i in measures_to_set_false if i not in ids_to_update]

    return measures_to_update, measures_to_set_false

def validate_candidate_measure(measure_to_register,
                               ncg=None,
                               parameters=[],
                               **kwargs):
    """
    Compute the matching distances, matching the reference measure to the measure
    originally registered to it. This is an inverse check from the original mathcing.
    In other words, the first registration registers A->B to find measure_to_register (B-naught).
    This func then matches B->A (B-prime) and computes the distance between B-naught and B-prime.

    Parameters
    ----------
    measure_to_register : dict
                          A dictionary containing information about the measure to validate, the
                          {keys: types} needed for this function are: 
                          {'id': int, 
                          'line': np.float, 
                          'sample': np.float,
                          'parameters_index': dict}

    ncg : obj
          A network candidate graph object

    geom_func : str
                The func to use to perform geometric matching

    match_func : str
                 The function to use to perform matching

    parameters : list
                 A list of matching parametrizations to test. Each entry results in
                 a subpixel registration attempt and then set of these results is
                 used ot ientify inliner and outlier parameter sets.

    Returns
    -------
    dists : list
            Of reprojection distances for each parameter set.
    """

    if not ncg.Session:
        raise BrokenPipeError('This func requires a database session from a NetworkCandidateGraph.')

    measure_to_register_id = measure_to_register['id']

    with ncg.session_scope() as session:
        # Get the measure to be registered
        measure = session.query(Measures).filter(Measures.id == measure_to_register_id).order_by(Measures.id).one()
        # Get the references measure
        point = measure.point
        reference_index = point.reference_index
        reference_measure = point.measures[reference_index]


        # Match the reference measure to the measure_to_register - this is the inverse of the first match attempt
        # Source is the image that we are seeking to validate, destination is the reference measure.
        # This is the inverse of other functions as this is a validator.

        source_imageid = measure.imageid
        source_image = session.query(Images).filter(Images.id == source_imageid).one()
        source_node = NetworkNode(node_id=source_imageid, image_path=source_image.path)
        source_node.parent = ncg

        destination_imageid = reference_measure.imageid
        destination_image = session.query(Images).filter(Images.id == destination_imageid).one()
        destination_node = NetworkNode(node_id=destination_imageid, image_path=destination_image.path)
        destination_node.parent = ncg

        sample = measure_to_register['sample']
        line = measure_to_register['line']

        log.info(f'Validating measure: {measure_to_register_id} on image: {source_imageid}')

        reference_roi = roi.Roi(source_node.geodata, 
                                sample, 
                                line, 
                                size_x=parameters[0]['match_kwargs']['image_size'][0],
                                size_y=parameters[0]['match_kwargs']['image_size'][1], 
                                buffer=10)
        moving_roi = roi.Roi(destination_node.geodata, 
                                reference_measure.sample, 
                                reference_measure.line, 
                                size_x=parameters[0]['match_kwargs']['template_size'][0],
                                size_y=parameters[0]['match_kwargs']['template_size'][1],
                                buffer=10)

        try:
            baseline_affine = estimate_local_affine(reference_roi, moving_roi)
        except:
            log.error('Unable to transform image to reference space. Likely too close to the edge of the non-reference image. Setting ignore=True')
            return [np.inf] * len(parameters)
        

        dists = []
        for i, parameter in enumerate(parameters):
            match_kwargs = parameter['match_kwargs']

            reference_roi = roi.Roi(source_node.geodata, 
                                    sample, 
                                    line, 
                                    size_x=match_kwargs['image_size'][0],
                                    size_y=match_kwargs['image_size'][1], 
                                    buffer=10)
            moving_roi = roi.Roi(destination_node.geodata, 
                                    reference_measure.sample, 
                                    reference_measure.line, 
                                    size_x=match_kwargs['template_size'][0],
                                    size_y=match_kwargs['template_size'][1],
                                    buffer=10)

            # Handle the exception where the clip can raise an index error if it is outside the image
            try:
                updated_affine, maxcorr, _ = subpixel_template(reference_roi,
                                                            moving_roi,
                                                            affine=baseline_affine)
            except:
                updated_affine = None

            if updated_affine is None:
                continue
            
            new_x, new_y = updated_affine([reference_measure.sample, 
                                           reference_measure.line])[0]
            
            dist = np.sqrt((new_y - reference_measure.line) ** 2 +\
                           (new_x - reference_measure.sample) ** 2)
            log.info(f'Validating using parameter set {i}. Reprojection distance: {dist}. Metric: {maxcorr}')
            dists.append(dist)
        return dists

def smart_register_point(pointid, parameters=[], shared_kwargs={}, valid_reprojection_distance=1.1, ncg=None, Session=None):
    """
    The entry func for the smart subpixel registration code. This is the user
    side API func for subpixel registering a point using the smart matcher.

    This function runs multiple rounds of subpixel registration on a point
    using 'subpixel_register_point_smart', checks for a consensus from the
    subpixel registration results, and validates the new location by inverting
    the matching direction. This function writes to the database and outputs
    the updated and ignored measures for logging purposes.

    This func writes to the databse. The returns are for logging and
    debugging convenience.

    Parameters
    ----------
    pointid : int or object
              The id of the point to register or a point object from which the
              id is accessed.

    parameters : list
                 A list of dict subpixel registration kwargs, {template_size: (x,x), image_size: (y,y)}

    shared_kwargs : dict
                    of kwargs passed to the subpixel matcher that are shared between all of the parameter sets

    ncg : obj
          A network candidate graph object

    Session : obj
              An optional sqlalchemy Session factory

    valid_reprojection_distance : float
                                  measures matched from the moving image to the reference image with a 
                                  distance less than this value in pixels are considered valid. Default: 1.1 

    Returns
    -------
    measures_to_update : list
                         of measures (dicts) to be updated to subpixel accuracy

    measures_to_set_false : list
                            of meaure ids to be ignored beause theu fail the consensus
                            building approach

    """
    if isinstance(pointid, Points):
        pointid = pointid.id
    measure_results = subpixel_register_point_smart(pointid, ncg=ncg, parameters=parameters, **shared_kwargs)
    measures_to_update, measures_to_set_false = decider(measure_results)

    log.info(f'Found {len(measures_to_update)} measures that found subpixel registration consensus. Running validation now...')
    # Validate that the new position has consensus
    for measure in measures_to_update:
        reprojection_distances = validate_candidate_measure(measure, parameters=parameters, ncg=ncg, **shared_kwargs)
        if np.sum(np.array(reprojection_distances) < valid_reprojection_distance) < 2:
            log.info(f"Measure {measure['id']} failed validation. Setting ignore=True for this measure.")
            measures_to_set_false.append(measure['id'])

    for measure in measures_to_update:
        measure['_id'] = measure.pop('id', None)

    # Update the measures that passed registration
    with ncg.engine.connect() as conn:
        if measures_to_update:
            stmt = Measures.__table__.update().\
                                    where(Measures.__table__.c.id == bindparam('_id')).\
                                    values({'weight':bindparam('weight'),
                                            'measureIgnore':bindparam('ignore'),
                                            'templateMetric':bindparam('template_metric'),
                                            'templateShift':bindparam('template_shift'),
                                            'line': bindparam('line'),
                                            'sample':bindparam('sample'),
                                            'ChooserName':bindparam('choosername')})
            resp = conn.execute(
                stmt, measures_to_update
            )
        if measures_to_set_false:
            measures_to_set_false = [{'_id':i} for i in measures_to_set_false]
            # Set ignore=True measures that failed
            stmt = Measures.__table__.update().\
                                    where(Measures.__table__.c.id == bindparam('_id')).\
                                    values({'measureIgnore':True,
                                            'ChooserName':shared_kwargs['chooser']})
            resp = conn.execute(
                stmt, measures_to_set_false
            )
    log.info(f'Updated measures: {json.dumps(measures_to_update, indent=2, cls=JsonEncoder)}')
    log.info(f'Ignoring measures: {measures_to_set_false}')

    return measures_to_update, measures_to_set_false


def mutual_information_match(moving_roi,
                             reference_roi, 
                             affine=tf.AffineTransform(), 
                             subpixel_size=3,
                             func=None, **kwargs):
    """
    Applies the mutual information matcher function over a search image using a
    defined template


    Parameters
    ----------
    moving_roi : roi 
                 The input search template used to 'query' the destination
                 image

    reference_roi : roi
              The image or sub-image to be searched

    subpixel_size : int
                    Subpixel area size to search for the center of mass
                    calculation

    func : function
           Function object to be used to compute the histogram comparison

    Returns
    -------
    new_affine :AffineTransform
                The affine transformation

    max_corr : float
               The strength of the correlation in the range [0, 4].

    corr_map : ndarray
               Map of corrilation coefficients when comparing the template to
               locations within the search area
    """
    reference_roi.clip()
    moving_roi.clip(affine=affine)

    moving_image = moving_roi.clipped_array
    reference_template = reference_roi.clipped_array

    if func == None:
        func = mutual_information

    image_size = moving_image.shape
    template_size = reference_template.shape

    y_diff = image_size[0] - template_size[0]
    x_diff = image_size[1] - template_size[1]

    max_corr = -np.inf
    corr_map = np.zeros((y_diff+1, x_diff+1))
    for i in range(y_diff+1):
        for j in range(x_diff+1):
            sub_image = moving_image[i:i+template_size[1],  # y
                                j:j+template_size[0]]  # x
            corr = func(sub_image, reference_template, **kwargs)
            if corr > max_corr:
                max_corr = corr
            corr_map[i, j] = corr

    y, x = np.unravel_index(np.argmax(corr_map, axis=None), corr_map.shape)

    upper = int(2 + floor(subpixel_size / 2))
    lower = upper - 1
    area = corr_map[y-lower:y+upper,
                    x-lower:x+upper]

    # Compute the y, x shift (subpixel) using scipys center_of_mass function
    cmass  = center_of_mass(area)
    if area.shape != (subpixel_size + 2, subpixel_size + 2):
        return  None, 0, None
        

    subpixel_y_shift = subpixel_size - 1 - cmass[0]
    subpixel_x_shift = subpixel_size - 1 - cmass[1]
    y = abs(y - (corr_map.shape[1])/2)
    x = abs(x - (corr_map.shape[0])/2)
    y += subpixel_y_shift
    x += subpixel_x_shift
    new_affine = tf.AffineTransform(translation=(-x, -y))
    return new_affine, np.max(max_corr), corr_map


import numpy as np


def response(row, edge):
    """
    Suppression function that converts 'response' into 'strength'
    """
    return row['response']


def correlation(row, edge):
    """
    Suppression function that converts 'correlation' into 'strength'
    """
    return row['correlation']


def distance(row, edge):
    """
    Suppression function that converts 'distance' into 'strength'
    """
    return 1 / row['distance']


def error(row, edge):
    """
    Suppression function that takes the reprojection error
    in a fundamental matrix as the inverse strength.
    """
    key = row.name
    try:
        return 1 / edge.fundamental_matrix.error.loc[key]
    except:
        return np.NaN
